from unittest.mock import MagicMock

import ipyvuetify as vw

import solara


def test_switch():
    on_value = MagicMock()
    el = solara.Switch(label="label", value=True, on_value=on_value)
    _, rc = solara.render(el, handle_error=False)
    switch = rc.find(vw.Switch)
    switch.widget.v_model = False
    assert on_value.call_count == 1
    switch.widget.v_model = False
    assert on_value.call_count == 1
    switch.widget.v_model = True
    assert on_value.call_count == 2
    rc.close()



"""# use_dark_effective"""

import solara
import solara.autorouting
import solara.lab
from solara.website.utils import apidoc

from . import NoPage

title = "use_dark_effective"
Page = NoPage
__doc__ += apidoc(solara.lab.use_dark_effective)  # type: ignore



"""
# Button

"""

import solara
from solara.website.utils import apidoc


@solara.component
def Page():
    count, set_count = solara.use_state(0)

    def increment():
        set_count(count + 1)

    with solara.VBox() as main:
        with solara.HBox():
            solara.Button(label=f"Clicked {count} times", on_click=increment, icon_name="mdi-thumb-up")
    return main


__doc__ += apidoc(solara.Button.f)  # type: ignore



"""# Title"""

from typing import Optional, cast

import solara
from solara.website.utils import apidoc


@solara.component
def Page():
    title = solara.use_reactive(cast(Optional[str], "Custom title!"))

    solara.ToggleButtonsSingle(value=title, values=[None, "Custom title!", "Different custom title"])

    if title is not None:
        # if the title is not set in a child component, the parent's title will be used
        with solara.Head():
            # title should always occur inside a Head component
            solara.Title(title)
        solara.Info(f"Your browser tab title should say {title}", classes=["mt-4"])
    else:
        solara.Warning("If no title is set, the parent title is used.", classes=["mt-4"])


__doc__ += apidoc(solara.Title.f)  # type: ignore



from typing import Callable, Union

import reacton
import reacton.ipyvuetify as v

import solara


@reacton.value_component(bool)
def Checkbox(
    *,
    label=None,
    value: Union[bool, solara.Reactive[bool]] = True,
    on_value: Callable[[bool], None] = None,
    disabled=False,
    style: str = None,
):
    """A checkbox is a widget that allows the user to toggle a boolean state.

    Basic examples

    ```solara
    import solara

    turbo_boost = solara.reactive(True)

    @solara.component
    def Page():
        checkbox = solara.Checkbox(label="Turbo boost", value=turbo_boost)
        if turbo_boost.value:
            solara.Success("Turbo boost is on")
        else:
            solara.Warning("Turbo boost is off, you might want to turn it on")
    ```


    ## Arguments

     * `label`: The label to display next to the checkbox.
     * `value`: The current value of the checkbox (True or False).
     * `on_value`: A callback that is called when the checkbox is toggled.
     * `disabled`: If True, the checkbox is disabled and cannot be used.
     * `style`: A string of CSS styles to apply to the checkbox.
    """
    reactive_value = solara.use_reactive(value, on_value)
    del value, on_value
    children = []
    if label is not None:
        children = [label]
    return v.Checkbox(label=label, v_model=reactive_value.value, on_v_model=reactive_value.set, disabled=disabled, style_=style, children=children)



"""
# Row
"""

import solara
import solara.lab
from solara.website.utils import apidoc

gap_size = solara.reactive("12px")
justify = solara.reactive("space-around")


@solara.component
def Page():
    with solara.Card("Row demo") as main:
        with solara.Column():
            solara.Text("Justify:")
            solara.ToggleButtonsSingle(justify, values=["start", "center", "end", "space-around", "space-between", "space-evenly"])
            solara.Select(
                label="Gap size",
                values=["0px", "4px", "8px", "12px", "16px", "20px", "24px"],
            ).connect(gap_size)
        with solara.Row(gap=gap_size.value, justify=justify.value):
            colors = "green red orange brown yellow pink".split()
            for color in colors:
                solara.Button(label="Solara", color=color)
    return main


__doc__ += apidoc(solara.Row.f)  # type: ignore



from typing import Dict, List, Optional, cast

from solara.toestand import Reactive

headers: Reactive[Optional[Dict[str, List[str]]]] = Reactive(cast(Optional[Dict[str, List[str]]], None), key="solara.lab.headers")



"""# Todo application

Demonstrates the use of reactive variables in combinations with dataclasses.

With solara we can get a type safe view onto a field in a dataclass, pydantic model, or
attr object.

This is using the experimental `solara.lab.Ref` class, which is not yet part of the
official API.


```python
import dataclasses
import solara
from solara.lab import Ref

@dataclasses.dataclass(frozen=True)
class TodoItem:
    text: str
    done: bool

todo_item = solara.reactive(TodoItem("Buy milk", False))

# now text is a reactive variable that is always in sync with todo_item.text
text = Ref(todo_item.fields.text)


# we can now modify the reactive text variable
# and see its value reflect in the todo_item
text.value = "Buy skimmed milk"
assert todo_item.value.text == "Buy skimmed milk"

# Or, the other way around
todo_item.value = TodoItem("Buy whole milk", False)
assert text.value == "Buy whole milk"
```

Apart from dataclasses, pydantic models etc, we also supports dictionaries and lists.

```python
todo_items = solara.reactive([TodoItem("Buy milk", False), TodoItem("Buy eggs", False)])
todo_item_eggs = Ref(todo_items.fields[1])
todo_item_eggs.value = TodoItem("Buy eggs", True)
assert todo_items.value[1].done == True

# However, if a list becomes shorter, and the valid index is now out of range, the
# reactive variables will act as if it is "not connected", and will not trigger
# updates anymore. Accessing the value will raise an IndexError.

todo_items.value = [TodoItem("Buy milk", False)]
# anyone listening to todo_item_eggs will *not* be notified.
try:
    value = todo_item_eggs.value
except IndexError:
    print("this is expected")
else:
    raise AssertionError("Expected an IndexError")
```

"""

import dataclasses
from typing import Callable

import reacton.ipyvuetify as v

import solara
from solara.lab.toestand import Ref


# our model for a todo item, immutable/frozen avoids common bugs
@dataclasses.dataclass(frozen=True)
class TodoItem:
    text: str
    done: bool


@solara.component
def TodoEdit(todo_item: solara.Reactive[TodoItem], on_delete: Callable[[], None], on_close: Callable[[], None]):
    """Takes a reactive todo item and allows editing it. Will not modify the original item until 'save' is clicked."""
    copy = solara.use_reactive(todo_item.value)

    def save():
        todo_item.value = copy.value
        on_close()

    with solara.Card("Edit", margin=0):
        solara.InputText(label="", value=Ref(copy.fields.text))
        with solara.CardActions():
            v.Spacer()
            solara.Button("Save", icon_name="mdi-content-save", on_click=save, outlined=True, text=True)
            solara.Button("Close", icon_name="mdi-window-close", on_click=on_close, outlined=True, text=True)
            solara.Button("Delete", icon_name="mdi-delete", on_click=on_delete, outlined=True, text=True)


@solara.component
def TodoListItem(todo_item: solara.Reactive[TodoItem], on_delete: Callable[[TodoItem], None]):
    """Displays a single todo item, modifications are done 'in place'.

    For demonstration purposes, we allow editing the item in a dialog as well.
    This will not modify the original item until 'save' is clicked.
    """
    edit, set_edit = solara.use_state(False)
    with v.ListItem():
        solara.Button(icon_name="mdi-delete", icon=True, on_click=lambda: on_delete(todo_item.value))
        solara.Checkbox(value=Ref(todo_item.fields.done))  # , color="success")
        solara.InputText(label="", value=Ref(todo_item.fields.text))
        solara.Button(icon_name="mdi-pencil", icon=True, on_click=lambda: set_edit(True))
        with v.Dialog(v_model=edit, persistent=True, max_width="500px", on_v_model=set_edit):
            if edit:  # 'reset' the component state on open/close

                def on_delete_in_edit():
                    on_delete(todo_item.value)
                    set_edit(False)

                TodoEdit(todo_item, on_delete=on_delete_in_edit, on_close=lambda: set_edit(False))


@solara.component
def TodoNew(on_new: Callable[[TodoItem], None]):
    """Component that managed entering new todo items"""
    new_text, set_new_text = solara.use_state("")
    text_field = v.TextField(v_model=new_text, on_v_model=set_new_text, label="Enter a new todo item")

    def create_new_item(*ignore_args):
        if not new_text:
            return
        new_item = TodoItem(text=new_text, done=False)
        on_new(new_item)
        # reset text
        set_new_text("")

    v.use_event(text_field, "keydown.enter", create_new_item)
    return text_field


initial_items = [
    TodoItem("Learn Solara", done=True),
    TodoItem("Write cool apps", done=False),
    TodoItem("Relax", done=False),
]


# We store out reactive state, and our logic in a class for organization
# purposes, but this is not required.
# Note that all the above components do not refer to this class, but only
# to do the Todo items.
# This means all above components are reusable, and can be used in other
# places, while the components below use 'application'/'global' state.
# They are not suited for reuse.


class State:
    todos = solara.reactive(initial_items)

    @staticmethod
    def on_new(item: TodoItem):
        State.todos.value = [item] + State.todos.value

    @staticmethod
    def on_delete(item: TodoItem):
        new_items = list(State.todos.value)
        new_items.remove(item)
        State.todos.value = new_items


@solara.component
def TodoStatus():
    """Status of our todo list"""
    items = State.todos.value
    count = len(items)
    items_done = [item for item in items if item.done]
    count_done = len(items_done)

    if count != count_done:
        with solara.Row():
            percent = count_done / count * 100
            solara.ProgressLinear(value=percent)
        with solara.Row():
            solara.Text(f"Remaining: {count - count_done}")
            solara.Text(f"Completed: {count_done}")
    else:
        solara.Success("All done, awesome!", dense=True)


@solara.component
def Page():
    with solara.Card("Todo list", style="min-width: 500px"):
        TodoNew(on_new=State.on_new)
        if State.todos.value:
            TodoStatus()
            for index, item in enumerate(State.todos.value):
                todo_item = Ref(State.todos.fields[index])
                TodoListItem(todo_item, on_delete=State.on_delete)
        else:
            solara.Info("No todo items, enter some text above, and hit enter")



import pytest
import contextlib
import solara.settings

import solara
from solara.validate_hooks import HookValidationError
import solara.validate_hooks


@contextlib.contextmanager
def hook_check_raise():
    prev = solara.settings.main.check_hooks
    solara.settings.main.check_hooks = "error"
    try:
        yield
    finally:
        solara.settings.main.check_hooks = prev


@contextlib.contextmanager
def hook_check_warn():
    prev = solara.settings.main.check_hooks
    solara.settings.main.check_hooks = "warn"
    try:
        yield
    finally:
        solara.settings.main.check_hooks = prev


def test_hooks_noqa_regex():
    regex = solara.validate_hooks.noqa_pattern
    assert regex.match("foo") is None
    assert regex.match("lala # noqa").groups() == (None, None)  # type: ignore
    assert regex.match("# noqa: SH102").groups() == ("SH102", None)  # type: ignore
    assert regex.match("# noqa: SH102, SH103").groups() == ("SH102", "SH103")  # type: ignore


def test_line_to_noqa():
    all = set(solara.validate_hooks.noqa_code_to_cause.values())
    assert solara.validate_hooks.line_to_noqa("def foo():") is None
    assert solara.validate_hooks.line_to_noqa("def foo():  # noqa") == all
    assert solara.validate_hooks.line_to_noqa("def foo():  # noqa: SH102") == {solara.validate_hooks.InvalidReactivityCause.CONDITIONAL_USE}
    assert solara.validate_hooks.line_to_noqa("def foo():  # noqa: SH102, SH103") == {
        solara.validate_hooks.InvalidReactivityCause.CONDITIONAL_USE,
        solara.validate_hooks.InvalidReactivityCause.LOOP_USE,
    }


def test_hook_use_invalid_loop():
    with hook_check_raise(), pytest.raises(HookValidationError):

        @solara.component
        def Page():
            for i in range(10):
                solara.use_state(1)
            solara.Text("Done")

    with hook_check_warn(), pytest.warns(UserWarning):

        @solara.component
        def Page2():
            for i in range(10):
                solara.use_state(1)
            solara.Text("Done")

    @solara.component
    def Page3():  # noqa: SH103
        for i in range(10):
            solara.use_state(1)
        solara.Text("Done")


def test_hook_validate_ast_checks():
    # a Call on a Subscript failed
    @solara.component
    def Page4():
        funcs = [lambda x: x]
        funcs[0](1)


def test_hook_use_invalid_conditional():
    with hook_check_warn(), pytest.warns(UserWarning):

        @solara.component
        def Page():
            if 1 > 3:
                solara.use_state(1)
            solara.Text("Done")

    @solara.component
    def Page2():
        if 1 > 3:
            solara.use_state(1)  # noqa: SH102
        solara.Text("Done")


def test_hook_use_early_return():
    with hook_check_raise(), pytest.raises(HookValidationError):

        @solara.component
        def Page():
            return
            solara.use_state(1)

    @solara.component
    def Page2():  # noqa: SH101
        return
        solara.use_state(1)

    @solara.component
    def Page3(  # noqa: SH101
        mul=1,
        ti=2,
        line=3,
        should="be triggered",
        by="having many and long",
        arguments="otherwise",
        out="linting will correct it",
    ):
        return
        solara.use_state(1)


def test_hook_use_nested_function():
    # sometimes we know that the use of a hook is stable, even when in a loop

    with hook_check_raise(), pytest.raises(HookValidationError):
        # single line
        @solara.component
        def Page():
            def inner_function():
                solara.use_state(1)

            inner_function()

    @solara.component
    def Page2():  # noqa: SH104
        def inner_function():
            solara.use_state(1)

        inner_function()


def test_hook_use_in_try():
    # sometimes we know that the use of a hook is stable, even when in a loop

    with hook_check_raise(), pytest.raises(HookValidationError):

        @solara.component
        def Page():
            try:
                solara.use_state(1)
            except ValueError:
                pass

    @solara.component
    def Page2():  # noqa: SH106
        try:
            solara.use_state(1)
        except ValueError:
            pass


def test_hook_use_invalid_assign():
    # sometimes we know that the use of a hook is stable, even when in a loop

    with hook_check_raise(), pytest.raises(HookValidationError):

        @solara.component
        def Page():
            alias = solara.use_state
            if 1 < 3:
                alias(1)



import io
from typing import Any, List

import solara
from solara.alias import rw


@solara.component
def FigureMatplotlib(
    figure,
    dependencies: List[Any] = None,
    format: str = "svg",
    **kwargs,
):
    """Display a matplotlib figure.


    ## Example

    ```solara
    import solara
    from matplotlib.figure import Figure

    @solara.component
    def Page():
        fig = Figure()
        ax = fig.subplots()
        ax.plot([1, 2, 3], [1, 4, 9])
        return solara.FigureMatplotlib(fig)
    ```

    When running under solara-server, we by default configure the same 'inline' backend as in the Jupyter notebook.

    For performance reasons, you might want to pass in a list of dependencies that indicate when
    the figure changed, to avoid re-rendering it on every render.

    ## Example using pyplot

    Note that it is also possible to use the pyplot interface, but be sure to close the figure not to leak memory.

    ```solara
    import solara
    import matplotlib.pyplot as plt

    @solara.component
    def Page():
        plt.figure()
        plt.plot([1, 2, 3], [1, 4, 9])
        plt.show()
        plt.close()
    ```

    Note that the figure is not the same size using the pyplot interface, due to the default figure size being different.


    ## Arguments

     * `figure`: Matplotlib figure.
     * `dependencies`: List of dependencies to watch for changes, if None, will convert the figure to a static image on each render.
     * `format`: The image format to to convert the Matplotlib figure to (png, jpg, svg, etc.)
     * `kwargs`: Additional arguments to passed to figure.savefig
    """

    def make_image():
        f = io.BytesIO()
        figure.savefig(f, format=format, **kwargs)
        return f.getvalue()

    value = solara.use_memo(make_image, dependencies)
    # mime type name is different from format name of matplotlib
    format_mime = format
    if format_mime == "svg":
        format_mime = "svg+xml"
    return rw.Image(value=value, format=format_mime)



"""# Echarts"""

import solara
from solara.website.utils import apidoc

options = {
    "bars": {
        "title": {"text": "ECharts Getting Started Example"},
        "tooltip": {},
        "legend": {"data": ["sales"]},
        "xAxis": {"type": "category"},
        "yAxis": {},
        "emphasis": {"itemStyle": {"shadowBlur": 10, "shadowOffsetX": 0, "shadowColor": "rgba(0, 0, 0, 0.5)"}},
        "series": [
            {
                "name": "sales",
                "type": "bar",
                "data": [
                    {"name": "Shirts", "value": 5},
                    {"name": "Cardigans", "value": 20},
                    {"name": "Chiffons", "value": 36},
                    {"name": "Pants", "value": 10},
                    {"name": "Heels", "value": 10},
                    {"name": "Socks", "value": 20},
                ],
                "universalTransition": True,
            }
        ],
    },
    "pie": {
        "title": {"text": "ECharts Getting Started Example"},
        "tooltip": {},
        "legend": {"data": ["sales"]},
        "series": [
            {
                "name": "sales",
                "type": "pie",
                "radius": [0, "50%"],
                "data": [
                    {"name": "Shirts", "value": 5},
                    {"name": "Cardigans", "value": 20},
                    {"name": "Chiffons", "value": 36},
                    {"name": "Pants", "value": 10},
                    {"name": "Heels", "value": 10},
                    {"name": "Socks", "value": 20},
                ],
                "universalTransition": True,
            }
        ],
    },
}


@solara.component
def Page():
    option, set_option = solara.use_state("bars")
    click_data, set_click_data = solara.use_state(None)
    mouseover_data, set_mouseover_data = solara.use_state(None)
    mouseout_data, set_mouseout_data = solara.use_state(None)

    with solara.VBox() as main:
        with solara.Card("Echarts"):
            with solara.ToggleButtonsSingle("bars", on_value=set_option):
                solara.Button("bars")
                solara.Button("pie")
            solara.FigureEcharts(
                option=options[option], on_click=set_click_data, on_mouseover=set_mouseover_data, on_mouseout=set_mouseout_data, responsive=True
            )
        with solara.Card("Event data"):
            solara.Markdown(f"**Click data**: {click_data}")
            solara.Markdown(f"**Mouseover data**: {mouseover_data}")
            solara.Markdown(f"**Mouseout data**: {mouseout_data}")

    return main


__doc__ += apidoc(solara.FigureEcharts.f)  # type: ignore



"""
# Overview
Click on one of the items on the left.
"""

import solara
from solara.website.components import CategoryLayout, Gallery

_title = "API"


@solara.component
def Page(route_external=None):
    Gallery(route_external)


@solara.component
def Layout(children=[]):
    CategoryLayout(children=children)



import asyncio
import json
import logging
import math
import os
from pathlib import Path
import sys
import threading
import typing
from typing import Any, Dict, List, Optional, Set, Union, cast
from uuid import uuid4
import warnings

import anyio
import starlette.websockets
import uvicorn.server
import websockets.legacy.http
import websockets.exceptions

from solara.server.utils import path_is_child_of

try:
    import solara_enterprise

    del solara_enterprise
    has_solara_enterprise = True
except ImportError:
    has_solara_enterprise = False
if has_solara_enterprise and sys.version_info[:2] > (3, 6):
    has_auth_support = True
    from solara_enterprise.auth.middleware import MutateDetectSessionMiddleware
    from solara_enterprise.auth.starlette import (
        AuthBackend,
        authorize,
        get_user,
        login,
        logout,
    )
else:
    has_auth_support = False

from starlette.applications import Starlette
from starlette.exceptions import HTTPException
from starlette.middleware import Middleware
from starlette.middleware.authentication import AuthenticationMiddleware
from starlette.middleware.gzip import GZipMiddleware
from starlette.requests import HTTPConnection, Request
from starlette.responses import HTMLResponse, JSONResponse, RedirectResponse, Response
from starlette.routing import Mount, Route, WebSocketRoute
from starlette.staticfiles import StaticFiles
from starlette.types import Receive, Scope, Send

import solara
import solara.settings
from solara.server.threaded import ServerBase

from . import app as appmod
from . import kernel_context, server, settings, telemetry, websocket
from .cdn_helper import cdn_url_path, get_path

os.environ["SERVER_SOFTWARE"] = "solara/" + str(solara.__version__)
limiter: Optional[anyio.CapacityLimiter] = None
lock = threading.Lock()


def _ensure_limiter():
    # in older anyios (<4) the limiter can only be created in an async context
    # so we call this in a starlette handler
    global limiter
    if limiter is None:
        with lock:
            if limiter is None:
                limiter = anyio.CapacityLimiter(settings.kernel.max_count if settings.kernel.max_count is not None else math.inf)


logger = logging.getLogger("solara.server.fastapi")
# if we add these to the router, the server_test does not run (404's)
prefix = ""

# The limit for starlette's http traffic should come from h11's DEFAULT_MAX_INCOMPLETE_EVENT_SIZE=16kb
# In practice, testing with 132kb cookies (server_test.py:test_large_cookie) seems to work fine.
# For the websocket, the limit is set to 4kb till 10.4, see
#  * https://github.com/aaugustin/websockets/blob/10.4/src/websockets/legacy/http.py#L14
# Later releases should set this to 8kb. See
#  * https://github.com/aaugustin/websockets/commit/8ce4739b7efed3ac78b287da7fb5e537f78e72aa
#  * https://github.com/aaugustin/websockets/issues/743
# Since starlette seems to accept really large values for http, lets do the same for websockets
# An arbitrarily large value we settled on for now is 32kb
# If we don't do this, users with many cookies will fail to get a websocket connection.
ws_major_version = int(websockets.__version__.split(".")[0])
if ws_major_version >= 13:
    websockets.legacy.http.MAX_LINE_LENGTH = int(os.environ.get("WEBSOCKETS_MAX_LINE_LENGTH", str(1024 * 32)))  # type: ignore
else:
    websockets.legacy.http.MAX_LINE = 1024 * 32  # type: ignore


class WebsocketDebugInfo:
    lock = threading.Lock()
    attempts = 0
    connecting = 0
    open = 0
    closed = 0


class WebsocketWrapper(websocket.WebsocketWrapper):
    ws: starlette.websockets.WebSocket

    def __init__(self, ws: starlette.websockets.WebSocket, portal: Optional[anyio.from_thread.BlockingPortal]) -> None:
        self.ws = ws
        self.portal = portal
        self.to_send: List[Union[str, bytes]] = []
        # following https://docs.python.org/3/library/asyncio-task.html#asyncio.create_task
        # we store a strong reference
        self.tasks: Set[asyncio.Task] = set()
        self.event_loop = asyncio.get_event_loop()
        self._thread_id = threading.get_ident()
        if settings.main.experimental_performance:
            self.task = asyncio.ensure_future(self.process_messages_task())

    async def process_messages_task(self):
        while True:
            await asyncio.sleep(0.05)
            while len(self.to_send) > 0:
                first = self.to_send.pop(0)
                if isinstance(first, bytes):
                    await self._send_bytes_exc(first)
                else:
                    await self._send_text_exc(first)

    async def _send_bytes_exc(self, data: bytes):
        # make sures we catch the starlette/websockets specific exception
        # and re-raise it as a websocket.WebSocketDisconnect
        try:
            await self.ws.send_bytes(data)
        except RuntimeError as e:
            # starlette throws a RuntimeError once you call send after the connection is closed
            # or RuntimeError: Unexpected ASGI message 'websocket.send', after sending 'websocket.close' or response already completed.
            # from uvicorn.protocols.websockets.websockets_impl.py
            if "close message" in repr(e) or "websocket.close" in repr(e):
                raise websocket.WebSocketDisconnect() from e
            else:
                raise
        except (websockets.exceptions.ConnectionClosed, starlette.websockets.WebSocketDisconnect, RuntimeError) as e:
            raise websocket.WebSocketDisconnect() from e

    async def _send_text_exc(self, data: str):
        # make sures we catch the starlette/websockets specific exception
        # and re-raise it as a websocket.WebSocketDisconnect
        try:
            await self.ws.send_text(data)
        except RuntimeError as e:
            if "close message" in repr(e) or "websocket.close" in repr(e):
                raise websocket.WebSocketDisconnect() from e
            else:
                raise
        except (websockets.exceptions.ConnectionClosed, starlette.websockets.WebSocketDisconnect, RuntimeError) as e:
            raise websocket.WebSocketDisconnect() from e

    def close(self):
        async def _close_exc():
            try:
                await self.ws.close()
            except RuntimeError as e:
                if "close message" in repr(e) or "websocket.close" in repr(e):
                    raise websocket.WebSocketDisconnect() from e
                else:
                    raise
            except (websockets.exceptions.ConnectionClosed, starlette.websockets.WebSocketDisconnect, RuntimeError) as e:
                raise websocket.WebSocketDisconnect() from e

        if self.portal is None:
            asyncio.ensure_future(_close_exc())
        else:
            self.portal.call(_close_exc)

    def send_text(self, data: str) -> None:
        if self.portal is None:
            task = self.event_loop.create_task(self._send_text_exc(data))
            self.tasks.add(task)
            task.add_done_callback(self.tasks.discard)
        else:
            if settings.main.experimental_performance:
                self.to_send.append(data)
            else:
                if self._thread_id == threading.get_ident():
                    warnings.warn("""You are triggering a websocket send from the event loop thread.
Support for this is experimental, and to avoid this message, make sure you trigger updates
that trigger this from a different thread, e.g.:

from anyio import to_thread
await to_thread.run_sync(my_update)
""")
                    task = self.event_loop.create_task(self._send_text_exc(data))
                    self.tasks.add(task)
                    task.add_done_callback(self.tasks.discard)
                else:
                    self.portal.call(self._send_text_exc, data)

    def send_bytes(self, data: bytes) -> None:
        if self.portal is None:
            task = self.event_loop.create_task(self._send_bytes_exc(data))
            self.tasks.add(task)
            task.add_done_callback(self.tasks.discard)
        else:
            if settings.main.experimental_performance:
                self.to_send.append(data)
            else:
                if self._thread_id == threading.get_ident():
                    warnings.warn("""You are triggering a websocket send from the event loop thread.
Support for this is experimental, and to avoid this message, make sure you trigger updates
that trigger this from a different thread, e.g.:

from anyio import to_thread
await to_thread.run_sync(my_update)
""")
                    task = self.event_loop.create_task(self._send_bytes_exc(data))
                    self.tasks.add(task)
                    task.add_done_callback(self.tasks.discard)

                self.portal.call(self._send_bytes_exc, data)

    async def receive(self):
        if self.portal is None:
            message = await asyncio.ensure_future(self.ws.receive())
        else:
            if hasattr(self.portal, "start_task_soon"):
                # version 3+
                fut = self.portal.start_task_soon(self.ws.receive)
            else:
                fut = self.portal.spawn_task(self.ws.receive)

            message = await asyncio.wrap_future(fut)
        if message.get("text") is not None:
            return message["text"]
        elif message.get("bytes") is not None:
            return message["bytes"]
        elif message.get("type") == "websocket.disconnect":
            raise websocket.WebSocketDisconnect()
        else:
            raise RuntimeError(f"Unknown message type {message}")


class ServerStarlette(ServerBase):
    server: uvicorn.server.Server
    name = "starlette"

    def __init__(self, port: int, host: str = "localhost", starlette_app=None, **kwargs):
        super().__init__(port, host, **kwargs)
        self.app = starlette_app or app

    def has_started(self):
        return self.server.started

    def signal_stop(self):
        self.server.should_exit = True
        # this cause uvicorn to not wait for background tasks, e.g.:
        # <Task pending name='Task-55'
        #  coro=<WebSocketProtocol.run_asgi() running at
        #  /.../uvicorn/protocols/websockets/websockets_impl.py:184>
        # wait_for=<Future pending cb=[<TaskWakeupMethWrapper object at 0x16896aa00>()]>
        # cb=[WebSocketProtocol.on_task_complete()]>
        self.server.force_exit = True
        self.server.lifespan.should_exit = True

    def serve(self):
        from uvicorn.config import Config
        from uvicorn.server import Server

        if sys.version_info[:2] < (3, 7):
            # make python 3.6 work
            import asyncio

            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)

        # uvloop will trigger a: RuntimeError: There is no current event loop in thread 'fastapi-thread'
        config = Config(self.app, host=self.host, port=self.port, **self.kwargs, access_log=False, loop="asyncio")
        self.server = Server(config=config)
        self.started.set()
        self.server.run()


async def kernels(id):
    return JSONResponse({"name": "lala", "id": "dsa"})


async def kernel_connection(ws: starlette.websockets.WebSocket):
    _ensure_limiter()
    try:
        with WebsocketDebugInfo.lock:
            WebsocketDebugInfo.attempts += 1
            WebsocketDebugInfo.connecting += 1
        await _kernel_connection(ws)
    finally:
        with WebsocketDebugInfo.lock:
            WebsocketDebugInfo.closed += 1
            WebsocketDebugInfo.open -= 1


async def _kernel_connection(ws: starlette.websockets.WebSocket):
    session_id = ws.cookies.get(server.COOKIE_KEY_SESSION_ID)

    if settings.oauth.private and not has_auth_support:
        raise RuntimeError("SOLARA_OAUTH_PRIVATE requires solara-enterprise")
    if has_auth_support and "session" in ws.scope:
        user = get_user(ws)
        if user is None and settings.oauth.private:
            await ws.accept()
            logger.error("app is private, requires login")
            await ws.close(code=1008, reason="app is private, requires login")
            return
    else:
        user = None

    if not session_id:
        logger.warning("no session cookie")
        session_id = "session-id-cookie-unavailable:" + str(uuid4())
    # we use the jupyter session_id query parameter as the key/id
    # for a page scope.
    page_id = ws.query_params["session_id"]
    if not page_id:
        logger.error("no page_id")
    kernel_id = ws.path_params["kernel_id"]
    if not kernel_id:
        logger.error("no kernel_id")
        await ws.close()
        return
    logger.info("Solara kernel requested for session_id=%s kernel_id=%s", session_id, kernel_id)
    await ws.accept()
    with WebsocketDebugInfo.lock:
        WebsocketDebugInfo.connecting -= 1
        WebsocketDebugInfo.open += 1

    async def run(ws_wrapper: WebsocketWrapper):
        if kernel_context.async_context_id is not None:
            kernel_context.async_context_id.set(uuid4().hex)
        assert session_id is not None
        assert kernel_id is not None
        telemetry.connection_open(session_id)
        headers_dict: Dict[str, List[str]] = {}
        for k, v in ws.headers.items():
            if k not in headers_dict.keys():
                headers_dict[k] = [v]
            else:
                headers_dict[k].append(v)
        await server.app_loop(ws_wrapper, ws.cookies, headers_dict, session_id, kernel_id, page_id, user)

    def websocket_thread_runner(ws_wrapper: WebsocketWrapper, portal: anyio.from_thread.BlockingPortal):
        async def run_wrapper():
            try:
                await run(ws_wrapper)
            except:  # noqa
                if portal is not None:
                    await portal.stop(cancel_remaining=True)
                raise
            finally:
                telemetry.connection_close(session_id)

        # sometimes throws: RuntimeError: Already running asyncio in this thread
        anyio.run(run_wrapper)  # type: ignore

    # this portal allows us to sync call the websocket calls from this current event loop we are in
    # each websocket however, is handled from a separate thread
    try:
        if settings.kernel.threaded:
            async with anyio.from_thread.BlockingPortal() as portal:
                ws_wrapper = WebsocketWrapper(ws, portal)
                thread_return = anyio.to_thread.run_sync(websocket_thread_runner, ws_wrapper, portal, limiter=limiter)  # type: ignore
                await thread_return
        else:
            ws_wrapper = WebsocketWrapper(ws, None)
            await run(ws_wrapper)
    finally:
        if settings.main.experimental_performance:
            try:
                ws_wrapper.task.cancel()
            except:  # noqa
                logger.exception("error cancelling websocket task")
        try:
            await ws.close()
        except:  # noqa
            pass


def close(request: Request):
    kernel_id = request.path_params["kernel_id"]
    page_id = request.query_params["session_id"]
    context = kernel_context.contexts.get(kernel_id, None)
    if context is not None:
        context.page_close(page_id)
    response = HTMLResponse(content="", status_code=200)
    return response


async def root(request: Request, fullpath: str = ""):
    forwarded_host = request.headers.get("x-forwarded-host")
    forwarded_proto = request.headers.get("x-forwarded-proto")
    host = request.headers.get("host")
    if forwarded_proto and forwarded_proto != request.scope["scheme"]:
        warnings.warn(f"""Header x-forwarded-proto={forwarded_proto!r} does not match scheme={request.scope["scheme"]!r} as given by the asgi framework (probably uvicorn)

This might be a configuration mismatch behind a reverse proxy and can cause issues with redirect urls, and auth.

Most likely, you need to trust your reverse proxy server, see:
    https://solara.dev/documentation/getting_started/deploying/self-hosted

If you use uvicorn (the default when you use `solara run`), make sure you
configure the following environment variables for uvicorn correctly:
UVICORN_PROXY_HEADERS=1  # only needed for uvicorn < 0.10, since it is the default after 0.10
FORWARDED_ALLOW_IPS="127.0.0.1"  # 127.0.0.1 is the default, replace this by the ip of the proxy server

Make sure you replace the IP with the correct IP of the reverse proxy server (instead of 127.0.0.1).

If you are sure that only the reverse proxy can reach the solara server, you can consider setting:
FORWARDED_ALLOW_IPS="*" # This can be a security risk, only use when you know what you are doing
""")
    if settings.oauth.private and not has_auth_support:
        raise RuntimeError("SOLARA_OAUTH_PRIVATE requires solara-enterprise")
    root_path = settings.main.root_path or ""
    if not settings.main.base_url:
        # Note:
        # starlette does not respect x-forwarded-host, and therefore
        # base_url and expected_origin below could be different
        # x-forwarded-host should only be considered if the same criteria in
        # uvicorn's ProxyHeadersMiddleware accepts x-forwarded-proto
        settings.main.base_url = str(request.base_url)
    # if not explicltly set,
    configured_root_path = settings.main.root_path
    scope = request.scope
    root_path_asgi = scope.get("route_root_path", scope.get("root_path", ""))
    if settings.main.root_path is None:
        # use the default root path from the app, which seems to also include the path
        # if we are mounted under a path
        root_path = root_path_asgi
        logger.debug("root_path: %s", root_path)
        # or use the script-name header, for instance when running under a reverse proxy
        script_name = request.headers.get("script-name")
        if script_name:
            logger.debug("override root_path using script-name header from %s to %s", root_path, script_name)
            root_path = script_name
        script_name = request.headers.get("x-script-name")
        if script_name:
            logger.debug("override root_path using x-script-name header from %s to %s", root_path, script_name)
            root_path = script_name
        settings.main.root_path = root_path

    # lets be flexible about the trailing slash
    # TODO: maybe we should be more strict about the trailing slash
    naked_root_path = settings.main.root_path.rstrip("/")
    naked_base_url = settings.main.base_url.rstrip("/")
    if not naked_base_url.endswith(naked_root_path):
        msg = f"""base url {naked_base_url!r} does not end with root path {naked_root_path!r}

This could be a configuration mismatch behind a reverse proxy and can cause issues with redirect urls, and auth.

See also https://solara.dev/documentation/getting_started/deploying/self-hosted
"""
        if "script-name" in request.headers:
            msg += f"""It looks like the reverse proxy sets the script-name header to {request.headers["script-name"]!r}
"""
        if "x-script-name" in request.headers:
            msg += f"""It looks like the reverse proxy sets the x-script-name header to {request.headers["x-script-name"]!r}
"""
        if configured_root_path:
            msg += f"""It looks like the root path was configured to {configured_root_path!r} in the settings
"""
        if root_path_asgi:
            msg += f"""It looks like the root path set by the asgi framework was configured to {root_path_asgi!r}
"""
        warnings.warn(msg)
    if host and forwarded_host and forwarded_proto:
        port = request.base_url.port
        ports = {"http": 80, "https": 443}
        expected_origin = f"{forwarded_proto}://{forwarded_host}"
        if port and port != ports[forwarded_proto]:
            expected_origin += f":{port}"
        starlette_origin = settings.main.base_url
        # strip off trailing / because we compare to the naked root path
        starlette_origin = starlette_origin.rstrip("/")
        if naked_root_path:
            # take off the root path
            starlette_origin = starlette_origin[: -len(naked_root_path)]
        if starlette_origin != expected_origin:
            warnings.warn(f"""Origin as determined by starlette ({starlette_origin!r}) does not match expected origin ({expected_origin!r}) based on x-forwarded-proto ({forwarded_proto!r}) and x-forwarded-host ({forwarded_host!r}) headers.

This might be a configuration mismatch behind a reverse proxy and can cause issues with redirect urls, and auth.
Most likely your proxy server sets the host header incorrectly (value for this request was {host!r})

See also https://solara.dev/documentation/getting_started/deploying/self-hosted
""")

    request_path = request.url.path
    if request_path.startswith(root_path):
        request_path = request_path[len(root_path) :]
    if request_path in server._redirects.keys():
        return RedirectResponse(server._redirects[request_path])

    content = server.read_root(request_path, root_path)
    if content is None:
        if settings.oauth.private and not request.user.is_authenticated:
            raise HTTPException(status_code=401, detail="Unauthorized")
        raise HTTPException(status_code=404, detail="Page not found by Solara router")

    if settings.oauth.private and not request.user.is_authenticated:
        from solara_enterprise.auth.starlette import login

        return await login(request)

    response = HTMLResponse(content=content)
    session_id = request.cookies.get(server.COOKIE_KEY_SESSION_ID) or str(uuid4())
    samesite = "lax"
    secure = False
    httponly = settings.session.http_only
    # we want samesite, so we can set a cookie when embedded in an iframe, such as on huggingface
    # however, samesite=none requires Secure https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Set-Cookie/SameSite
    # when hosted on the localhost domain we can always set the Secure flag
    # to allow samesite https://developer.mozilla.org/en-US/docs/Web/HTTP/Cookies#restrict_access_to_cookies
    if request.scope["scheme"] == "https" or request.headers.get("x-forwarded-proto", "http") == "https" or request.base_url.hostname == "localhost":
        samesite = "none"
        secure = True
    elif request.base_url.hostname != "localhost":
        warnings.warn(f"""Cookies with samesite=none require https, but according to the asgi framework, the scheme is {request.scope["scheme"]!r}
and the x-forwarded-proto header is {request.headers.get("x-forwarded-proto", "http")!r}. We will fallback to samesite=lax.

If you embed solara in an iframe, make sure you forward the x-forwarded-proto header correctly so that the session cookie can be set.

See https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Set-Cookie/SameSite for more information on samesite cookies.

Also check out the following Solara documentation:
 * https://solara.dev/documentation/getting_started/deploying/self-hosted
 * https://solara.dev/documentation/advanced/howto/embed
""")
    response.set_cookie(
        server.COOKIE_KEY_SESSION_ID,
        value=session_id,
        expires="Fri, 01 Jan 2038 00:00:00 GMT",
        samesite=samesite,  # type: ignore
        secure=secure,  # type: ignore
        httponly=httponly,  # type: ignore
    )  # type: ignore
    return response


class StaticFilesOptionalAuth(StaticFiles):
    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        conn = HTTPConnection(scope)
        if settings.oauth.private and not has_auth_support:
            raise RuntimeError("SOLARA_OAUTH_PRIVATE requires solara-enterprise")
        if has_auth_support and settings.oauth.private and not conn.user.is_authenticated:
            raise HTTPException(status_code=401, detail="Unauthorized")
        await super().__call__(scope, receive, send)


class StaticNbFiles(StaticFilesOptionalAuth):
    def get_directories(
        self,
        directory: Union[str, "os.PathLike[str]", None] = None,
        packages=None,  # type: ignore
    ) -> List[Union[str, "os.PathLike[str]"]]:
        return cast(List[Union[str, "os.PathLike[str]"]], server.nbextensions_directories)

    # follow symlinks
    # from https://github.com/encode/starlette/pull/1377/files
    def lookup_path(self, path: str) -> typing.Tuple[str, typing.Optional[os.stat_result]]:
        for directory in self.all_directories:
            directory = os.path.realpath(directory)
            original_path = os.path.join(directory, path)
            full_path = os.path.realpath(original_path)
            # return early if someone tries to access a file outside of the directory
            if not path_is_child_of(Path(original_path), Path(directory)):
                return "", None
            try:
                return full_path, os.stat(full_path)
            except (FileNotFoundError, NotADirectoryError):
                continue
        return "", None


class StaticPublic(StaticFilesOptionalAuth):
    def lookup_path(self, *args, **kwargs):
        self.all_directories = self.get_directories(None, None)
        return super().lookup_path(*args, **kwargs)

    def get_directories(
        self,
        directory: Union[str, "os.PathLike[str]", None] = None,
        packages=None,  # type: ignore
    ) -> List[Union[str, "os.PathLike[str]"]]:
        # we only know the .directory at runtime (after startup)
        # which means we cannot pass the directory to the StaticFiles constructor
        return cast(List[Union[str, "os.PathLike[str]"]], [app.directory.parent / "public" for app in appmod.apps.values()])


class StaticAssets(StaticFilesOptionalAuth):
    def lookup_path(self, *args, **kwargs):
        self.all_directories = self.get_directories(None, None)
        return super().lookup_path(*args, **kwargs)

    def get_directories(
        self,
        directory: Union[str, "os.PathLike[str]", None] = None,
        packages=None,  # type: ignore
    ) -> List[Union[str, "os.PathLike[str]"]]:
        # we only know the .directory at runtime (after startup)
        # which means we cannot pass the directory to the StaticFiles constructor
        directories = server.asset_directories()
        return cast(List[Union[str, "os.PathLike[str]"]], directories)


class StaticCdn(StaticFilesOptionalAuth):
    def lookup_path(self, path: str) -> typing.Tuple[str, typing.Optional[os.stat_result]]:
        try:
            full_path = str(get_path(settings.assets.proxy_cache_dir, path))
        except Exception:
            return "", None
        return full_path, os.stat(full_path)


def on_startup():
    appmod.ensure_apps_initialized()
    # TODO: configure and set max number of threads
    # see https://github.com/encode/starlette/issues/1724
    telemetry.server_start()


def on_shutdown():
    # shutdown all kernels
    for context in list(kernel_context.contexts.values()):
        try:
            context.close()
        except:  # noqa
            logger.exception("error closing kernel on shutdown")
    telemetry.server_stop()


def readyz(request: Request):
    json, status = server.readyz()
    return JSONResponse(json, status_code=status)


def _sanitize_for_json(value):
    """Convert infinite values to None for JSON serialization."""
    if isinstance(value, (int, float)) and math.isinf(value):
        return None
    return value


async def resourcez(request: Request):
    _ensure_limiter()
    assert limiter is not None
    data: Dict[str, Any] = {}
    verbose = request.query_params.get("verbose", None) is not None
    data["websockets"] = {
        "attempts": WebsocketDebugInfo.attempts,
        "connecting": WebsocketDebugInfo.connecting,
        "open": WebsocketDebugInfo.open,
        "closed": WebsocketDebugInfo.closed,
    }
    from . import patch

    data["threads"] = {
        "created": patch.ThreadDebugInfo.created,
        "running": patch.ThreadDebugInfo.running,
        "stopped": patch.ThreadDebugInfo.stopped,
        "active": threading.active_count(),
    }
    contexts = list(kernel_context.contexts.values())
    data["kernels"] = {
        "total": len(contexts),
        "has_connected": len([k for k in contexts if kernel_context.PageStatus.CONNECTED in k.page_status.values()]),
        "has_disconnected": len([k for k in contexts if kernel_context.PageStatus.DISCONNECTED in k.page_status.values()]),
        "has_closed": len([k for k in contexts if kernel_context.PageStatus.CLOSED in k.page_status.values()]),
        "limiter": {
            "total_tokens": _sanitize_for_json(limiter.total_tokens),
            "borrowed_tokens": _sanitize_for_json(limiter.borrowed_tokens),
            "available_tokens": _sanitize_for_json(limiter.available_tokens),
        },
    }
    default_limiter = anyio.to_thread.current_default_thread_limiter()
    data["anyio.to_thread.limiter"] = {
        "total_tokens": _sanitize_for_json(default_limiter.total_tokens),
        "borrowed_tokens": _sanitize_for_json(default_limiter.borrowed_tokens),
        "available_tokens": _sanitize_for_json(default_limiter.available_tokens),
    }
    if verbose:
        try:
            import psutil

            def expand(named_tuple):
                return {key: getattr(named_tuple, key) for key in named_tuple._fields}

            data["cpu"] = {}
            try:
                data["cpu"]["percent"] = psutil.cpu_percent()
            except Exception as e:
                data["cpu"]["percent"] = str(e)
            try:
                data["cpu"]["count"] = psutil.cpu_count()
            except Exception as e:
                data["cpu"]["count"] = str(e)
            try:
                data["cpu"]["times"] = expand(psutil.cpu_times())
                data["cpu"]["times"]["per_cpu"] = [expand(x) for x in psutil.cpu_times(percpu=True)]
            except Exception as e:
                data["cpu"]["times"] = str(e)
            try:
                data["cpu"]["times_percent"] = expand(psutil.cpu_times_percent())
                data["cpu"]["times_percent"]["per_cpu"] = [expand(x) for x in psutil.cpu_times_percent(percpu=True)]
            except Exception as e:
                data["cpu"]["times_percent"] = str(e)
            try:
                memory = psutil.virtual_memory()
            except Exception as e:
                data["memory"] = str(e)
            else:
                data["memory"] = {
                    "bytes": expand(memory),
                    "GB": {key: getattr(memory, key) / 1024**3 for key in memory._fields},
                }

        except ModuleNotFoundError:
            pass

    json_string = json.dumps(data, indent=2)
    return Response(content=json_string, media_type="application/json")


middleware = [
    Middleware(GZipMiddleware, minimum_size=1000),
]

if has_auth_support:
    middleware = [
        *middleware,
        Middleware(
            MutateDetectSessionMiddleware,
            secret_key=settings.session.secret_key,  # type: ignore
            session_cookie="solara-session",  # type: ignore
            https_only=settings.session.https_only,  # type: ignore
            same_site=settings.session.same_site,  # type: ignore
        ),
        Middleware(AuthenticationMiddleware, backend=AuthBackend()),
    ]

routes_auth = []
if has_auth_support:
    routes_auth = [
        Route("/_solara/auth/authorize", endpoint=authorize),  #
        Route("/_solara/auth/logout", endpoint=logout),
        Route("/_solara/auth/login", endpoint=login),
    ]
routes = [
    Route("/readyz", endpoint=readyz),
    Route("/resourcez", endpoint=resourcez),
    *routes_auth,
    Route("/jupyter/api/kernels/{id}", endpoint=kernels),
    WebSocketRoute("/jupyter/api/kernels/{kernel_id}/{name}", endpoint=kernel_connection),
    Route("/", endpoint=root),
    Route("/{fullpath}", endpoint=root),
    Route("/_solara/api/close/{kernel_id}", endpoint=close, methods=["POST"]),
    # only enable when the proxy is turned on, otherwise if the directory does not exists we will get an exception
    *([Mount(f"/{cdn_url_path}", app=StaticCdn(directory=settings.assets.proxy_cache_dir))] if solara.settings.assets.proxy else []),
    Mount(f"{prefix}/static/public", app=StaticPublic()),
    Mount(f"{prefix}/static/assets", app=StaticAssets()),
    Mount(f"{prefix}/jupyter/nbextensions", app=StaticNbFiles()),
    Mount(f"{prefix}/static", app=StaticFilesOptionalAuth(directory=server.solara_static)),
    Route("/{fullpath:path}", endpoint=root),
]

app = Starlette(routes=routes, on_startup=[on_startup], on_shutdown=[on_shutdown], middleware=middleware)

# Uncomment the lines below to test solara mouted under a subpath
# def myroot(request: Request):
#     return JSONResponse({"framework": "solara"})

# routes_test_sub = [Route("/", endpoint=myroot), Mount("/foo/", routes=routes)]
# app = Starlette(routes=routes_test_sub, on_startup=[on_startup], on_shutdown=[on_shutdown], middleware=middleware)



from unittest.mock import MagicMock

import ipyvuetify as vw

import solara


def test_checkbox():
    on_value = MagicMock()
    el = solara.Checkbox(label="label", value=True, on_value=on_value)
    box, rc = solara.render(el, handle_error=False)
    checkbox = rc.find(vw.Checkbox)
    checkbox.widget.v_model = False
    assert on_value.call_count == 1
    checkbox.widget.v_model = False
    assert on_value.call_count == 1
    rc.close()


def test_checkbox_no_callback_on_managed():
    on_value = MagicMock()
    el = solara.Checkbox(label="label", value=True, on_value=on_value)
    box, rc = solara.render(el, handle_error=False)
    assert on_value.call_count == 0

    # changing the value externally should *not* call the callback
    el2 = solara.Checkbox(label="label", value=False, on_value=on_value)
    rc.render(el2)
    assert on_value.call_count == 0



from typing import Callable, Dict, List, Union, cast

import yaml
import markdown
import mkdocs_pycafe
import pymdownx.superfences

import solara
from solara.components.markdown import formatter, _no_deep_copy_emojione


# We want to separate metadata from the markdown files before rendering them, which solara.Markdown doesn't support
@solara.component
def MarkdownWithMetadata(content: str, unsafe_solara_execute=True):
    cleanups = solara.use_ref(cast(List[Callable[[], None]], []))
    if "---" in content:
        pre_content, raw_metadata, post_content = content.split("---")
        metadata: Dict[str, Union[str, List[str]]] = yaml.safe_load(raw_metadata)

        if len(pre_content) == 0:
            content = post_content
        else:
            content = pre_content + post_content

        if "title" not in metadata.keys():
            metadata["title"] = content.split("#")[1].split("\n")[0]

        for key, value in metadata.items():
            if key == "title":
                solara.Title(value)
            elif ":" in key:
                solara.Meta(property=key, content=value)
            else:
                solara.Meta(name=key, content=value)

    def make_markdown_object():
        return markdown.Markdown(  # type: ignore
            extensions=[
                "pymdownx.highlight",
                "pymdownx.superfences",
                "pymdownx.emoji",
                "toc",  # so we get anchors for h1 h2 etc
                "tables",
            ],
            extension_configs={
                "pymdownx.emoji": {
                    "emoji_index": _no_deep_copy_emojione,
                },
                "pymdownx.superfences": {
                    "custom_fences": [
                        {
                            "name": "mermaid",
                            "class": "mermaid",
                            "format": pymdownx.superfences.fence_div_format,
                        },
                        {
                            "name": "solara",
                            "class": "",
                            "validator": mkdocs_pycafe.validator,
                            "format": mkdocs_pycafe.formatter(
                                type="solara", next_formatter=formatter(unsafe_solara_execute, cleanups.current), inside_last_div=False
                            ),
                        },
                        {
                            "name": "python",
                            "class": "highlight",
                            "validator": mkdocs_pycafe.validator,
                            "format": mkdocs_pycafe.formatter(
                                type="solara", next_formatter=formatter(unsafe_solara_execute, cleanups.current), inside_last_div=False
                            ),
                        },
                    ],
                },
            },
        )

    md_parser = solara.use_memo(make_markdown_object, dependencies=[unsafe_solara_execute])

    def cleanup_wrapper():
        def cleanup():
            for cleanup in cleanups.current:
                cleanup()

        return cleanup

    solara.use_effect(cleanup_wrapper, [])

    with solara.v.Html(
        tag="div",
        style_="display: flex; flex-direction: row; justify-content: center; gap: 15px; max-width: 90%; margin: 0 auto;",
        attributes={"id": "markdown-to-navigate"},
    ):
        solara.Markdown(content, unsafe_solara_execute=unsafe_solara_execute, style="flex-grow: 1; max-width: min(100%, 1024px);", md_parser=md_parser)
        MarkdownNavigation(id="markdown-to-navigate").key("markdown-nav" + str(hash(content)))


@solara.component_vue("markdown_nav.vue")
def MarkdownNavigation(id: str):
    pass



from typing import Callable, Optional

import ipyvue as vue
import ipyvuetify as v

import solara

routes = [
    solara.Route(path="/"),
    solara.Route(
        path="fruit",
        children=[
            solara.Route(path="kiwi"),
            solara.Route(path="banana"),
            solara.Route(path="apple"),
        ],
    ),
    solara.Route(
        path="blog",
        children=[
            solara.Route(path="/"),
            solara.Route(path="foo"),
            solara.Route(path="bar"),
        ],
    ),
    solara.Route(path="contact"),
]


def test_router():
    assert solara.routing.Router("", routes).path_routes == [routes[0]]
    assert solara.routing.Router("/fruit", routes).path_routes == [routes[1]]
    assert solara.routing.Router("/fruit/kiwi", routes).path_routes == [routes[1], routes[1].children[0]]

    assert solara.routing.Router("/fruit/pineapple", routes).path_routes == [routes[1]]
    assert solara.routing.Router("/fruit/apple", routes).path_routes == [routes[1], routes[1].children[-1]]

    assert solara.routing.Router("/blog", routes).path_routes == [routes[2], routes[2].children[0]]
    assert solara.routing.Router("/blog/doesnotexist", routes).path_routes == [routes[2]]
    assert solara.routing.Router("/blog/doesnotexist/more", routes).path_routes == [routes[2]]
    assert solara.routing.Router("/blog/foo", routes).path_routes == [routes[2], routes[2].children[1]]
    assert solara.routing.Router("/blog/foo/", routes).path_routes == [routes[2], routes[2].children[1]]

    assert solara.routing.Router("/doesnotexist", routes).path_routes == []

    assert solara.routing.Router("?a=1", routes).path_routes == [routes[0]]
    assert solara.routing.Router("?a=1", routes).search == "a=1"
    assert solara.routing.Router("/fruit?b=1&c=3", routes).path_routes == [routes[1]]

    # non-existing routes, as leafs are fine, since they can do 'subrouting'
    assert solara.routing.Router("/fruit/kiwi/sub", routes).path_routes == [routes[1], routes[1].children[0]]
    assert solara.routing.Router("/fruit/kiwi/sub", routes).possible_match

    # but if there are sublings, it should not exist
    assert solara.routing.Router("/fruit/chocolate", routes).path_routes == [routes[1]]
    assert not solara.routing.Router("/fruit/chocolate", routes).possible_match

    assert solara.routing.Router("/fruit/chocolate", routes).path_routes == [routes[1]]
    assert not solara.routing.Router("/fruit/chocolate/sub", routes).possible_match

    assert solara.routing.Router("/foo", routes).path_routes == []
    assert not solara.routing.Router("/foo", routes).possible_match
    assert solara.routing.Router("/foo/bar", routes).path_routes == []
    assert not solara.routing.Router("/foo/bar", routes).possible_match


def test_resolve_path_route():
    @solara.component
    def Test(route):
        return solara.Text(solara.resolve_path(route))

    @solara.component
    def Provider(path, route):
        # nonlocal set_path
        path, set_path = solara.use_state_or_update(path)
        solara.routing._location_context.provide(solara.routing._Location(path, set_path))
        solara.routing.router_context.provide(solara.routing.Router(path, routes=routes))
        return Test(route)

    container, rc = solara.render(Provider("/", routes[0]))
    assert rc._find(vue.Html).widget.children[0] == "/"

    container, rc = solara.render(Provider("/", routes[1]))
    assert rc._find(vue.Html).widget.children[0] == "/fruit"
    container, rc = solara.render(Provider("/", routes[1].children[1]))
    assert rc._find(vue.Html).widget.children[0] == "/fruit/banana"


def test_resolve_path_str():
    @solara.component
    def Test(path: str):
        solara.use_route()
        return solara.Text(solara.resolve_path(path))

    @solara.component
    def FruitProvider(path: str):
        path, set_path = solara.use_state_or_update(path)
        solara.routing._location_context.provide(solara.routing._Location(path, set_path))
        solara.routing.router_context.provide(solara.routing.Router("/fruit", routes=routes))
        solara.use_route()
        return Test(path)

    container, rc = solara.render(FruitProvider("kiwi"))
    assert rc._find(vue.Html).widget.children[0] == "/fruit/kiwi"
    container, rc = solara.render(FruitProvider("apple"))
    assert rc._find(vue.Html).widget.children[0] == "/fruit/apple"
    container, rc = solara.render(FruitProvider("/fruit/apple"))
    assert rc._find(vue.Html).widget.children[0] == "/fruit/apple"


def test_toggle_buttons_single():
    value: Optional[str] = None
    set_path: Callable[[str], None] = lambda x: None  # noqa: E731

    def set(value_):
        nonlocal value
        value = value_

    routes = [
        solara.Route(path="/"),
        solara.Route(
            path="fruit",
            children=[
                solara.Route(path="kiwi"),
                solara.Route(path="banana"),
                solara.Route(path="apple"),
            ],
        ),
    ]

    @solara.component
    def Test():
        route, routes = solara.use_route()
        if route is None:
            return solara.Button("Error!")
        assert solara.resolve_path(route.path) == f"/fruit/{route.path}"
        route_banana = solara.routing.find_route("banana")
        assert route_banana is not None
        assert route_banana.path == "banana"
        assert solara.resolve_path(route_banana.path) == "/fruit/banana"
        with solara.Link(solara.resolve_path(route_banana.path)) as main:
            solara.Button(route.path)
        return main

    @solara.component
    def Root():
        route, routes = solara.use_route()
        assert solara.resolve_path("fruit") == "/fruit"
        assert solara.resolve_path("fruit/banana") == "/fruit/banana"
        assert route in routes
        assert route is not None
        assert route.path in ["fruit"]
        return Test()

    @solara.component
    def Provider():
        nonlocal set_path
        path, set_path = solara.use_state("/fruit/banana")

        solara.routing._location_context.provide(solara.routing._Location(path, set_path))
        solara.routing.router_context.provide(solara.routing.Router(path, routes=routes))
        return Root()

    container, rc = solara.render(Provider(), handle_error=False)
    assert rc._find(v.Btn).widget.children[0] == "banana"
    set_path("/fruit/kiwi")
    assert rc._find(v.Btn).widget.children[0] == "kiwi"
    set_path("/fruit/wumpa")
    assert rc._find(v.Btn).widget.children[0] == "Error!"



import sys
from pathlib import Path

import ipyreact
import playwright
import pytest

import solara
import solara.server.patch

solara.server.patch.patch()


HERE = Path(__file__).parent


test_js_code = """
import * as React from "react";

export default function({value, setValue, ...children}) {
    return React.createElement("button", {onClick: () => setValue(value + 1), children: `clicked ${value || 0}`})
}
"""
ipyreact.define_module("solara-test", test_js_code)


@solara.component
def Page():
    value = solara.use_reactive(0)
    ipyreact.ValueWidget.element(_module="solara-test", value=0, on_value=value.set)


@solara.component
def PageDefineDuringRun():
    ipyreact.define_module("solara-test-dynamic", test_js_code)
    value = solara.use_reactive(0)
    ipyreact.ValueWidget.element(_module="solara-test-dynamic", value=0, on_value=value.set)


@pytest.mark.skipif(sys.version_info < (3, 7, 0), reason="ipyreact requires python 3.7 or higher")
@pytest.mark.parametrize("app", ["esm_test:Page", "esm_test:PageDefineDuringRun"])
def test_ipyreact(browser: playwright.sync_api.Browser, page_session: playwright.sync_api.Page, solara_server, solara_app, extra_include_path, app):
    with extra_include_path(HERE), solara_app(app):
        page_session.goto(solara_server.base_url)
        page_session.locator("text=clicked 0").first.click()
        page_session.locator("text=clicked 1").first.click()
        # now force a page reload / new kernel
        # which would fail if the module and import map widget
        # was shared between kernels
        page_session.goto(solara_server.base_url)
        page_session.locator("text=clicked 0").first.click()
        page_session.locator("text=clicked 1").first.click()



import threading
import typing
from typing import Any, Callable, List, Optional, Union, cast

import traitlets
from ipyvue import Template
from ipyvuetify.extra import FileInput
from ipywidgets import widget_serialization
from typing_extensions import TypedDict

import solara
import solara.hooks as hooks


class FileInfo(TypedDict):
    name: str
    size: int
    file_obj: typing.BinaryIO
    data: Optional[bytes]


class FileDropZone(FileInput):
    # override to narrow traitlet of FileInput
    template = traitlets.Instance(Template).tag(sync=True, **widget_serialization)
    template_file = (__file__, "file_drop.vue")
    items = traitlets.List(default_value=cast(List[Any], [])).tag(sync=True)
    label = traitlets.Unicode().tag(sync=True)
    multiple = traitlets.Bool(True).tag(sync=True)


@solara.component
def _FileDrop(
    label="Drop file(s) here",
    on_total_progress: Optional[Callable[[float], None]] = None,
    on_file: Optional[Callable[[Union[FileInfo, List[FileInfo]]], None]] = None,
    lazy: bool = True,
    multiple: bool = False,
):
    """Generic implementation used by FileDrop and FileDropMultiple.

    If multiple=True, multiple files can be uploaded.
    """

    file_info, set_file_info = solara.use_state(None)
    wired_files, set_wired_files = solara.use_state(cast(Optional[typing.List[FileInfo]], None))

    file_drop = FileDropZone.element(label=label, on_total_progress=on_total_progress, on_file_info=set_file_info, multiple=multiple)  # type: ignore

    def wire_files():
        if not file_info:
            return

        real = cast(FileDropZone, solara.get_widget(file_drop))

        # workaround for @observe being cleared
        real.version += 1
        real.reset_stats()

        set_wired_files(cast(typing.List[FileInfo], real.get_files()))

    solara.use_effect(wire_files, [file_info])

    def handle_file(cancel: threading.Event):
        if not wired_files:
            return
        if on_file:
            for i in range(len(wired_files)):
                if not lazy:
                    wired_files[i]["data"] = wired_files[i]["file_obj"].read()
                else:
                    wired_files[i]["data"] = None
            if multiple:
                on_file(wired_files)
            else:
                on_file(wired_files[0])

    result: solara.Result = hooks.use_thread(handle_file, [wired_files])
    if result.error:
        raise result.error

    return file_drop


@solara.component
def FileDrop(
    label="Drop file here",
    on_total_progress: Optional[Callable[[float], None]] = None,
    on_file: Optional[Callable[[FileInfo], None]] = None,
    lazy: bool = True,
):
    """Region a user can drop a file into for file uploading.

    If lazy=True, no file content will be loaded into memory,
    nor will any data be transferred by default.
    If lazy=False, file content will be loaded into memory and passed to the `on_file` callback via the `FileInfo.data` attribute.


    A file object is of the following argument type:
    ```python
    class FileInfo(typing.TypedDict):
        name: str  # file name
        size: int  # file size in bytes
        file_obj: typing.BinaryIO
        data: Optional[bytes]: bytes  # only present if lazy=False
    ```


    ## Arguments
     * `on_total_progress`: Will be called with the progress in % of the file upload.
     * `on_file`: Will be called with a `FileInfo` object, which contains the file `.name`, `.length` and a `.file_obj` object.
     * `lazy`: Whether to load the file contents into memory or not. If `False`,
        the file contents will be loaded into memory via the `.data` attribute of file object(s).

    ## Load into Pandas
    To load the data into a Pandas DF, set `lazy=False` and use `file['file_obj']` (be careful of memory)<br>
    You can run this directly in your Jupyter notebook

    ```python
    import io
    import pandas as pd
    import solara

    @solara.component
    def Page():
        def load_file_df(file):
            df = pd.read_csv(file["file_obj"])
            print("Loaded dataframe:")
            print(df)

        solara.FileDrop(label="Drop file to see dataframe!", on_file=load_file_df)

    ```

    """

    return _FileDrop(label=label, on_total_progress=on_total_progress, on_file=on_file, lazy=lazy, multiple=False)


@solara.component
def FileDropMultiple(
    label="Drop files here",
    on_total_progress: Optional[Callable[[float], None]] = None,
    on_file: Optional[Callable[[List[FileInfo]], None]] = None,
    lazy: bool = True,
):
    """Region a user can drop multiple files into for file uploading.

    Almost identical to `FileDrop` except that multiple files can be dropped and `on_file` is called
    with a list of `FileInfo` objects.

    ## Arguments
     * `on_total_progress`: Will be called with the progress in % of the file(s) upload.
     * `on_file`: Will be called with a `List[FileInfo]`.
        Each `FileInfo` contains the file `.name`, `.length`, `.file_obj` object, and `.data` attributes.
     * `lazy`: Whether to load the file contents into memory or not.

    """

    return _FileDrop(label=label, on_total_progress=on_total_progress, on_file=on_file, lazy=lazy, multiple=True)



import datetime as dt
from typing import Optional
from unittest.mock import MagicMock

import ipyvuetify as vw

import solara
from solara.lab import InputDate, InputDateRange

today = dt.date.today()
tomorrow = today + dt.timedelta(days=1)


def test_input_date_typing():
    def on_value(value: dt.date):
        pass

    InputDate(value=today, label="label", on_value=on_value)

    def on_value2(value: Optional[dt.date]):
        pass

    InputDate(value=today, label="label", on_value=on_value2)


def test_input_date():
    on_value = MagicMock()
    on_v_model = MagicMock()
    el = InputDate(value=today, label="label", on_value=on_value)
    box, rc = solara.render(el, handle_error=False)
    input = rc.find(vw.TextField)
    input.widget.observe(on_v_model, "v_model")

    input.widget.v_model = tomorrow.strftime("%Y/%m/%d")
    assert on_value.call_count == 1
    assert on_v_model.call_count == 1
    on_v_model.reset_mock()
    assert on_value.call_args[0][0] == tomorrow

    input.widget.v_model = ""
    assert on_value.call_count == 1
    assert on_v_model.call_count == 1
    assert on_value.call_args[0][0] == tomorrow
    assert input.widget.error
    assert input.widget.label == "label (Date cannot be empty)"

    input.widget.v_model = "2023/10/17"
    assert on_value.call_count == 2
    assert on_v_model.call_count == 2
    assert on_value.call_args[0][0] == dt.date(2023, 10, 17)
    assert not input.widget.error
    assert input.widget.v_model == dt.date(2023, 10, 17).strftime("%Y/%m/%d")


def test_input_date_optional():
    on_value = MagicMock()
    on_v_model = MagicMock()
    el = InputDate(value=today, label="label", on_value=on_value, optional=True)
    box, rc = solara.render(el, handle_error=False)
    input = rc.find(vw.TextField)
    input.widget.observe(on_v_model, "v_model")

    input.widget.v_model = tomorrow.strftime("%Y/%m/%d")
    assert on_value.call_count == 1
    assert on_v_model.call_count == 1
    assert on_value.call_args[0][0] == tomorrow

    input.widget.v_model = ""
    assert on_value.call_count == 2
    assert on_v_model.call_count == 2
    assert on_value.call_args[0][0] is None
    assert not input.widget.error

    input.widget.v_model = "2023/10/17"
    assert on_value.call_count == 3
    assert on_v_model.call_count == 3
    assert on_value.call_args[0][0] == dt.date(2023, 10, 17)
    assert not input.widget.error
    assert input.widget.v_model == dt.date(2023, 10, 17).strftime("%Y/%m/%d")


def test_input_date_incomplete_entry():
    on_value = MagicMock()
    on_v_model = MagicMock()

    @solara.component
    def Test():
        def update_value(value: dt.date):
            on_value(value)
            set_value(value)

        value, set_value = solara.use_state(today)

        InputDate(value=value, on_value=update_value, label="label")

    el = Test()
    box, rc = solara.render(el, handle_error=False)
    input = rc.find(vw.TextField)
    input.widget.observe(on_v_model, "v_model")

    input.widget.v_model = "2023/10"
    assert on_value.call_count == 0
    assert on_v_model.call_count == 1
    assert on_value.call_args is None
    assert input.widget.error
    assert input.widget.v_model == "2023/10"
    assert on_v_model.call_count == 1

    input.widget.v_model = "2023/10/17"
    assert on_value.call_count == 1
    assert on_v_model.call_count == 2
    # Without below assert mypy fails
    assert on_value.call_args is not None
    assert on_value.call_args[0][0] == dt.date(2023, 10, 17)
    assert not input.widget.error
    assert input.widget.v_model == dt.date(2023, 10, 17).strftime("%Y/%m/%d")


def test_input_date_range():
    on_value = MagicMock()
    on_v_model = MagicMock()
    el = InputDateRange(value=tuple([today, tomorrow]), label="label", on_value=on_value)
    box, rc = solara.render(el, handle_error=False)
    text_field = rc.find(vw.TextField)
    input = rc.find(vw.DatePicker)
    text_field.widget.observe(on_v_model, "v_model")

    assert input.widget.v_model == [today.strftime("%Y-%m-%d"), tomorrow.strftime("%Y-%m-%d")]
    input.widget.v_model = [tomorrow.strftime("%Y-%m-%d"), today.strftime("%Y-%m-%d")]
    assert on_value.call_count == 1
    assert on_v_model.call_count == 1
    on_v_model.reset_mock()
    assert on_value.call_args[0][0] == tuple([tomorrow, today])

    input.widget.v_model = ["2023-10-17", None]
    assert on_value.call_count == 2
    # We enter no date, so on_v_model does not get called
    assert on_v_model.call_count == 1
    assert on_value.call_args[0][0] == tuple([dt.date(2023, 10, 17), None])
    assert text_field.widget.error
    assert text_field.widget.label == "label (Please select two dates)"

    input.widget.v_model = ["2023-10-17", tomorrow.strftime("%Y-%m-%d")]
    assert on_value.call_count == 3
    assert on_v_model.call_count == 2
    assert on_value.call_args[0][0] == tuple([dt.date(2023, 10, 17), tomorrow])
    assert not text_field.widget.error
    assert text_field.widget.v_model == f"{dt.date(2023, 10, 17).strftime('%Y/%m/%d')} - {tomorrow.strftime('%Y/%m/%d')}"



from typing import Callable

import ipyvuetify
import traitlets

import solara


class MarkdownEditorWidget(ipyvuetify.VuetifyTemplate):
    template_file = (__file__, "markdown_editor.vue")

    value = traitlets.Unicode("").tag(sync=True)
    height = traitlets.Unicode("180px").tag(sync=True)
    cdn = traitlets.Unicode(None, allow_none=True).tag(sync=True)

    @traitlets.default("cdn")
    def _cdn(self):
        import solara.settings

        if not solara.settings.assets.proxy:
            return solara.settings.assets.cdn


@solara.component
def MarkdownEditor(value: str = "", on_value: Callable[[str], None] = None):
    """WYSIWYG (visual) Markdown editor.

    ## Arguments

    * value: Markdown text
    * on_value: Callback function that is called when the text is changed
    """
    return MarkdownEditorWidget.element(value=value, on_value=on_value)



from pathlib import Path

import solara
from solara.website.components.sidebar import Sidebar

_title = "Documentation"

HERE = Path(__file__).parent
Sidebar = Sidebar

route_order = ["/", "getting_started", "examples", "components", "api", "advanced", "faq"]


@solara.component_vue(str(HERE.parent.parent / "components" / "algolia_api.vue"))
def Algolia():
    pass


@solara.component
def Page():
    # show a gallery of all the api pages
    router = solara.use_router()
    route_current = router.path_routes[-2]
    if route_current.path == "documentation":
        with solara.Column(style={"width": "100%", "max-width": "1024px"}, gap="40px"):
            with solara.Column(style={"width": "100%", "align-items": "center"}, gap="40px"):
                with solara.Column(gap="20px", classes=["docs-card-container"], style={"max-width": "80%", "padding-top": "64px"}, align="stretch"):
                    for route in route_current.children:
                        if route.path in ["/", "advanced", "faq"]:
                            continue
                        with solara.Link("/documentation/" + route.path):
                            with solara.Row(
                                classes=["docs-card"],
                                style={
                                    "background-color": f"var({'--docs-color-grey' if route.path != 'getting_started' else '--color-primary'})",
                                    "align-items": "center",
                                },
                            ):
                                solara.HTML(tag="h2", unsafe_innerHTML=route.label, style={"color": "white", "display": "block", "flex-grow": "1"})
                                solara.v.Icon(children=["mdi-arrow-right"], color="var(--color-grey-light)", x_large=True, class_="docs-card-icon")
                with solara.Column(gap="10px", style={"flex-wrap": "wrap", "max-width": "80%"}):
                    solara.HTML(tag="h2", unsafe_innerHTML="How to use our documentation:", style={"padding": "1.5rem"})
                    solara.Markdown(
                        """
* [Getting Started](/documentation/getting_started) - Learn how to install Solara and get started with building your app.
    Also includes tutorials for you to get a hang of Solara workflow.
* [Examples](/documentation/examples) - More complex and real world applicable examples of Solara apps.
    For even more complexity you can see the [Showcase](/showcase) page.
* [Components](/documentation/components) - All the components that are available in Solara.
* [API](/documentation/api) - All the API functions that are available in Solara. Importantly, this includes routing and hooks.
* [Advanced](/documentation/advanced) - Advanced topics like associated and underlying libraries.
    If a component you would like to use is not available in Solara, you can use the underlying library directly.
                    """
                    )
                    solara.HTML(tag="h2", unsafe_innerHTML="Also Check Out", style={"padding": "1.5rem"})
                    with solara.Row(gap="20px", style={"flex-wrap": "wrap", "row-gap": "20px", "align-items": "center", "padding-left": "24px"}):
                        with solara.v.Html(tag="a", attributes={"href": "https://discord.solara.dev", "target": "_blank"}):
                            with solara.Div(classes=["social-logo-container"], style={"background-color": "var(--docs-social-discord)"}):
                                solara.v.Html(tag="img", attributes={"src": "/static/public/social/discord.svg"}, style_="height: 1.5rem; width: auto;")
                        solara.Text("We use discord to provide support and answer questions there actively.")
                    with solara.Row(gap="20px", style={"flex-wrap": "wrap", "row-gap": "20px", "align-items": "center", "padding-left": "24px"}):
                        with solara.v.Html(tag="a", attributes={"href": "https://github.com/widgetti/solara", "target": "_blank"}):
                            with solara.Div(classes=["social-logo-container"], style={"background-color": "var(--docs-social-github)"}):
                                solara.v.Html(tag="img", attributes={"src": "/static/public/social/github.svg"}, style_="height: 1.5rem; width: auto;")
                        solara.Text("Search for solutions on Github issues, or report bugs.")
                    with solara.Row(gap="20px", style={"flex-wrap": "wrap", "row-gap": "20px", "align-items": "center", "padding-left": "24px"}):
                        with solara.v.Html(tag="a", attributes={"href": "https://twitter.com/solara_dev", "target": "_blank"}):
                            with solara.Div(classes=["social-logo-container"], style={"background-color": "var(--docs-social-twitter)"}):
                                solara.v.Html(tag="img", attributes={"src": "/static/public/social/twitter.svg"}, style_="height: 1.5rem; width: auto;")
                        solara.Text("Get announcements about Solara features, showcases, and events!.")


@solara.component
def Layout(children=[]):
    route_current, all_routes = solara.use_route()
    if route_current is None:
        return solara.Error("Page not found")

    if route_current.path == "/":
        return Page()
    else:
        with solara.Column(
            align="center", children=children, style={"padding": "0 0 40px 0", "max-width": "90%"}
        ):  # 40px bottom margin for "runs on Solara" text
            pass


@solara.component
def Overview():
    pass



"""
# on_kernel_start

Run a function when a virtual kernel (re)starts and optionally run a cleanup function on shutdown.

```python
def on_kernel_start(f: Callable[[], Optional[Callable[[], None]]]) -> Callable[[], None]:
    ...
```

`f` will be called on each virtual kernel (re)start. This (usually) happens each time a browser tab connects to the server
[see solara server for more details](https://solara.dev/docs/understanding/solara-server).
The (optional) function returned by `f` will be called on kernel shutdown.

Note that the cleanup functions are called in reverse order with respect to the order in which they were registered
(e.g. the cleanup function of the last call to `on_kernel_start` will be called first on kernel shutdown).

The return value of on_kernel_start is a cleanup function that will remove the callback from the list of callbacks to be called on kernel start.

During hot reload, the callbacks that are added from scripts or modules that will be reloaded will be removed before the app is loaded
again. This can cause the order of the callbacks to be different than at first run.

## Example

```
import solara
import solara.lab


@solara.lab.on_kernel_start
def on_kernel_start():
    id = solara.get_kernel_id()
    print("Kernel started", id)
    def cleanup():
        print("Kernel stopped", id)
    return cleanup  # this function will be called on kernel shutdown
```

"""

from solara.website.components import NoPage

title = "on_kernel_start"
Page = NoPage



from typing import Optional
from unittest.mock import MagicMock

import ipyvuetify as vw

import solara


def test_input_int_typing():
    def on_value(value: int):
        pass

    solara.InputInt("label", 42, optional=False, on_value=on_value)

    def on_value2(value: Optional[int]):
        pass

    solara.InputInt("label", 42, optional=True, on_value=on_value2)


def test_input_int_optional():
    on_value = MagicMock()
    el = solara.InputInt("label", 42, optional=True, on_value=on_value)
    box, rc = solara.render(el, handle_error=False)
    input = rc.find(vw.TextField)
    input.widget.v_model = "43"
    assert on_value.call_count == 0
    input.widget.fire_event("focusout")
    assert on_value.call_count == 1
    assert on_value.call_args[0][0] == 43

    input.widget.v_model = ""
    assert on_value.call_count == 1
    input.widget.fire_event("focusout")
    assert on_value.call_count == 2
    assert on_value.call_args[0][0] is None
    rc.close()

    el = solara.InputInt("label", None, optional=True, on_value=on_value)
    box, rc = solara.render(el, handle_error=False)
    input = rc.find(vw.TextField)
    assert input.widget.v_model is None

    rc.render(solara.InputInt("label", 1, optional=True, on_value=on_value))
    assert input.widget.v_model == "1"


def test_input_int():
    on_value = MagicMock()
    on_v_model = MagicMock()
    el = solara.InputInt("label", 42, optional=False, on_value=on_value)
    box, rc = solara.render(el, handle_error=False)
    input = rc.find(vw.TextField)
    input.widget.observe(on_v_model, "v_model")

    input.widget.v_model = "43"
    assert on_value.call_count == 0
    assert on_v_model.call_count == 1
    on_v_model.reset_mock()
    input.widget.fire_event("focusout")
    assert on_value.call_count == 1
    assert on_value.call_args[0][0] == 43
    assert on_v_model.call_count == 0

    input.widget.v_model = ""
    assert on_value.call_count == 1
    assert on_v_model.call_count == 1
    on_v_model.reset_mock()
    input.widget.fire_event("focusout")
    assert on_value.call_count == 1
    assert on_value.call_args[0][0] == 43
    assert input.widget.error
    assert input.widget.label == "label (Value cannot be empty)"
    assert on_v_model.call_count == 0

    input.widget.v_model = "44"
    assert on_value.call_count == 1
    assert on_v_model.call_count == 1
    input.widget.fire_event("focusout")
    assert on_v_model.call_count == 1
    assert on_value.call_count == 2
    assert on_value.call_args[0][0] == 44
    assert not input.widget.error
    assert input.widget.label == "label"


def test_input_int_managed():
    on_value = MagicMock()
    on_v_model = MagicMock()

    @solara.component
    def Test():
        def update_value(value: int):
            on_value(value)
            set_value(value)

        value, set_value = solara.use_state(42)
        solara.InputInt("label", value, optional=False, on_value=update_value)

    el = Test()
    box, rc = solara.render(el, handle_error=False)
    input = rc.find(vw.TextField)
    input.widget.observe(on_v_model, "v_model")

    input.widget.v_model = "1e3"
    assert on_value.call_count == 0
    assert on_v_model.call_count == 1
    input.widget.fire_event("focusout")
    assert on_value.call_count == 0
    assert input.widget.error
    assert input.widget.label == "label (Value must be an integer)"
    assert input.widget.v_model == "1e3"
    assert on_v_model.call_count == 1

    input.widget.v_model = "1"
    input.widget.fire_event("focusout")
    assert on_value.call_count == 1
    assert on_v_model.call_count == 2
    assert not input.widget.error

    input.widget.v_model = "1.1"
    assert on_value.call_count == 1
    assert on_v_model.call_count == 3
    assert not input.widget.error
    input.widget.fire_event("focusout")
    # no change
    assert on_value.call_count == 1
    assert input.widget.error
    assert input.widget.label == "label (Value must be an integer)"
    assert input.widget.v_model == "1.1"


def test_input_int_external():
    on_value = MagicMock()
    on_v_model_input = MagicMock()
    on_v_model_slider = MagicMock()

    @solara.component
    def Test():
        def update_value(value: int):
            on_value(value)
            set_value(value)

        value, set_value = solara.use_state(42)
        solara.InputInt("label", value, optional=False, on_value=update_value)
        solara.IntSlider("label", value, min=0, max=100, step=1, on_value=update_value)

    el = Test()
    box, rc = solara.render(el, handle_error=False)
    input = rc.find(vw.TextField)
    slider = rc.find(vw.Slider)
    input.widget.observe(on_v_model_input, "v_model")
    slider.widget.observe(on_v_model_slider, "v_model")

    input.widget.v_model = "43"
    assert on_value.call_count == 0
    assert on_v_model_input.call_count == 1
    assert on_v_model_slider.call_count == 0
    input.widget.fire_event("focusout")
    assert on_value.call_count == 1
    assert on_v_model_input.call_count == 1
    assert on_v_model_slider.call_count == 1
    assert on_value.call_args[0][0] == 43
    assert not input.widget.error

    slider.widget.v_model = 44
    assert on_value.call_count == 2
    assert on_v_model_slider.call_count == 2
    assert on_v_model_input.call_count == 2
    assert not input.widget.error
    assert input.widget.v_model == "44"

    input.widget.v_model = "1.1"
    assert on_value.call_count == 2
    assert on_v_model_input.call_count == 3
    assert on_v_model_slider.call_count == 2
    assert not input.widget.error
    input.widget.fire_event("focusout")
    # No call to on_value or slider v_model, because the value is invalid
    assert on_value.call_count == 2
    assert on_v_model_slider.call_count == 2
    assert on_v_model_input.call_count == 3
    assert input.widget.error
    slider.widget.v_model = 42
    assert on_value.call_count == 3
    assert on_v_model_slider.call_count == 3
    assert on_v_model_input.call_count == 4
    assert not input.widget.error


def test_input_float_managed():
    on_value = MagicMock()
    on_v_model = MagicMock()

    @solara.component
    def Test():
        def update_value(value: float):
            on_value(value)
            set_value(value)

        value, set_value = solara.use_state(42.1)
        solara.InputFloat("label", value, optional=False, on_value=update_value)

    el = Test()
    box, rc = solara.render(el, handle_error=False)
    input = rc.find(vw.TextField)
    input.widget.observe(on_v_model, "v_model")

    input.widget.v_model = "1.1e3"
    assert on_value.call_count == 0
    assert on_v_model.call_count == 1
    input.widget.fire_event("focusout")
    assert on_value.call_count == 1
    assert on_v_model.call_count == 1
    assert on_value.call_args[0][0] == 1100
    assert not input.widget.error
    assert input.widget.label == "label"
    assert input.widget.v_model == "1.1e3"

    input.widget.v_model = "1,1e3"
    assert on_value.call_count == 1
    assert on_v_model.call_count == 2
    input.widget.fire_event("focusout")
    assert on_value.call_count == 1
    assert on_v_model.call_count == 2
    # assert on_value.call_args[0][0] == 1100
    assert not input.widget.error
    assert input.widget.label == "label"
    assert input.widget.v_model == "1,1e3"

    input.widget.v_model = "1.1e0"
    assert on_value.call_count == 1
    assert on_v_model.call_count == 3
    input.widget.fire_event("focusout")
    assert on_value.call_count == 2
    assert on_v_model.call_count == 3
    assert on_value.call_args[0][0] == 1.1
    assert not input.widget.error
    assert input.widget.label == "label"
    assert input.widget.v_model == "1.1e0"

    # same value
    input.widget.v_model = "1.1"
    assert on_v_model.call_count == 4
    assert on_value.call_count == 2
    input.widget.fire_event("focusout")
    assert on_v_model.call_count == 4
    # no change
    assert on_value.call_count == 2
    assert not input.widget.error
    assert input.widget.label == "label"
    assert input.widget.v_model == "1.1"



from typing import List
import solara


@solara.component
def BreadCrumbs():
    router = solara.use_router()
    routes = router.path_routes

    with solara.Row(style={"align-items": "center", "flex-wrap": "wrap"}) as main:
        for i, route in enumerate(routes):
            if i == len(routes) - 1:
                solara.Text(route.label or route.path, style={"color": "var(--color-text-fade)"})
            else:
                with solara.Link(solara.resolve_path(route), style={"color": "var(--color-text-fade)"}):
                    solara.Text(route.label or route.path)
            if i != len(routes) - 1:
                solara.Text("/", style={"font-size": "1.5rem", "color": "var(--color-text-fade)"})
    return main


def _resolve_path_to_route(path_to_find: List[str], all_routes: List[solara.Route], routes: List[solara.Route] = []):
    if len(path_to_find) == 0:
        return routes
    for route in all_routes:
        if path_to_find[0] == route.path:
            routes += [route]
            return _resolve_path_to_route(path_to_find[1:], route.children, routes)



import solara
from solara.website.components import Gallery, MarkdownWithMetadata

title = "Examples"

pycafe_projects = [
    "chatbot",
    "tokenizer",
]


@solara.component
def Page(route_external=None):
    Gallery(route_external)


@solara.component
def Layout(children):
    # TODO: this is using a private API, what is the best way to do this?
    # we want to 'eat' the whole route for the current level, and the level below
    # for example the utilities directory. But if an example does routing, we don't want
    # to take on that route.
    router = solara.use_router()
    route_current = router.path_routes[-1]
    # route_current, all_routes = solara.use_route()

    if route_current is None:
        return solara.Error("Page not found")
    module = route_current.module
    assert module is not None
    github_url = solara.util.github_url(module.__file__)

    if route_current.path == "fullscreen":
        with solara.Padding(4, children=children):
            pass
    else:
        with solara.Column(align="center", style={"max-width": "100%"}):
            doc = module.__doc__
            if doc:
                with solara.Column():
                    MarkdownWithMetadata(doc)
            with solara.Column(style={"max-width": "min(100%, 1024px)", "width": "100%"}):
                if route_current.path != "/":
                    solara.Button("View source code on GitHub", icon_name="mdi-github-circle", href=github_url, class_="ma-2", target="_blank", text=True)
                if route_current.path in pycafe_projects:
                    pycafe_url = f"https://py.cafe/solara/{route_current.path}"
                    solara.Button(
                        "Run this example on PyCafe", icon_name="mdi-coffee-to-go-outline", href=pycafe_url, class_="ma-2", target="_blank", text=True
                    )
                if not hasattr(module, "Page"):
                    solara.Error(f"No Page component found in {module}")
                else:
                    with solara.Div(children=children):
                        pass



import os
from typing import Optional

import solara.util

from .minisettings import BaseSettings, Field
from .util import get_solara_home

try:
    import dotenv
except ImportError:
    pass
else:
    dotenv.load_dotenv()


home = get_solara_home()
if not home.exists():
    try:
        home.mkdir(parents=True, exist_ok=True)
    except OSError:
        pass  # can fail in for instance docker when $HOME is not set/writable


class Cache(BaseSettings):
    type: str = Field("memory", env="SOLARA_CACHE", title="Type of cache, e.g. 'memory', 'disk', 'redis', or a multilevel cache, e.g. 'memory,disk'")
    disk_max_size: str = Field("10GB", title="Maximum size for'disk' cache , e.g. 10GB, 500MB")
    memory_max_size: str = Field("1GB", title="Maximum size for 'memory-size' cache, e.g. 10GB, 500MB")
    memory_max_items: int = Field(128, title="Maximum number of items for 'memory' cache")
    clear: bool = Field(False, title="Clear the cache on startup, only applies to disk and redis caches")
    path: Optional[str] = Field(
        os.path.join(home, "cache"), env="SOLARA_CACHE_PATH", title="Storage location for 'disk' cache. Defaults to `${SOLARA_HOME}/cache`"
    )

    class Config:
        env_prefix = "solara_cache_"
        case_sensitive = False
        env_file = ".env"


# in colab or vscode there is not solara cdn proxy available
_should_use_proxy = not (solara.util.is_running_in_colab() or solara.util.is_running_in_vscode() or solara.util.is_running_in_voila())


class Assets(BaseSettings):
    cdn: str = "https://cdn.jsdelivr.net/npm/"
    proxy: bool = _should_use_proxy

    class Config:
        env_prefix = "solara_assets_"
        case_sensitive = False
        env_file = ".env"


class MainSettings(BaseSettings):
    check_hooks: str = "warn"
    allow_reactive_boolean: bool = True
    # TODO: also change default_container in solara/components/__init__.py
    default_container: Optional[str] = "Column"

    class Config:
        env_prefix = "solara_"
        case_sensitive = False
        env_file = ".env"


class Storage(BaseSettings):
    mutation_detection: Optional[bool] = None  # True/False, or None to auto determine
    factory: str = "solara.toestand.default_storage"

    def get_factory(self):
        return solara.util.import_item(self.factory)

    class Config:
        env_prefix = "solara_storage_"
        case_sensitive = False
        env_file = ".env"


assets: Assets = Assets()
cache: Cache = Cache()
main = MainSettings()
storage = Storage()

if main.check_hooks not in ["off", "warn", "raise"]:
    raise ValueError(f"Invalid value for check_hooks: {main.check_hooks}, expected one of ['off', 'warn', 'raise']")



from typing import Callable, Optional, Union, List
from .input import use_change
import solara
from solara.alias import rv as v


@solara.component
def InputTextArea(
    label: str,
    value: Union[str, solara.Reactive[str]] = "",
    on_value: Callable[[str], None] = None,
    disabled: bool = False,
    continuous_update: bool = False,
    update_events: List[str] = ["focusout"],
    error: Union[bool, str] = False,
    message: Optional[str] = None,
    auto_grow: bool = True,
    rows: int = 5,
    dense: bool = False,
    hide_details: Union[str, bool] = "auto",
    placeholder: Optional[str] = None,
    prefix: Optional[str] = None,
    suffix: Optional[str] = None,
):
    r"""Free form text area input.

    ### Basic example:

    ```solara
    import solara

    text = solara.reactive("Hello\nWorld\n!!!")
    continuous_update = solara.reactive(True)

    @solara.component
    def Page():
        solara.Checkbox(label="Continuous update", value=continuous_update)
        solara.InputTextArea("Enter some text", value=text, continuous_update=continuous_update.value)
        with solara.Row():
            solara.Button("Clear", on_click=lambda: text.set(""))
            solara.Button("Reset", on_click=lambda: text.set("Hello\nWorld\n!!!"))
        solara.Markdown(f"**You entered**: {text.value}")
    ```


    ## Arguments

    * `label`: Label to display next to the slider.
    * `value`: The currently entered value.
    * `on_value`: Callback to call when the value changes.
    * `disabled`: Whether the input is disabled.
    * `continuous_update`: Whether to call the `on_value` callback on every change or only when the input loses focus or the enter key is pressed.
    * `update_events`: A list of events that should trigger `on_value`. If continuous update is enabled, this will effectively be ignored,
        since updates will happen every change.
    * `auto_grow`: Whether the text area auto grows with more text.
    * `rows`: Number of empty rows to display.
    * `error`: If truthy, show the input as having an error (in red). If a string is passed, it will be shown as the error message.
    * `message`: Message to show below the input. If `error` is a string, this will be ignored.
    * `classes`: List of CSS classes to apply to the input.
    * `style`: CSS style to apply to the input.
    * `dense`: Reduces the input height.
    * `hide_details`: Hides hint and validation errors. When set to 'auto', messages will be rendered only if there's a message (hint, error message, counter value etc) to display.
    * `placeholder`: Sets the input's placeholder text.
    * `prefix`: Displays prefix text.
    * `suffix`: Displays suffix text.
    """
    reactive_value = solara.use_reactive(value, on_value)
    del value, on_value

    def set_value_cast(value):
        reactive_value.value = str(value)

    def on_v_model(value):
        if continuous_update:
            set_value_cast(value)

    messages = []
    if error and isinstance(error, str):
        messages.append(error)
    elif message:
        messages.append(message)
    text_area = v.Textarea(
        v_model=reactive_value.value,
        on_v_model=on_v_model,
        label=label,
        disabled=disabled,
        error=bool(error),
        messages=messages,
        solo=True,
        hide_details=hide_details,
        outlined=True,
        rows=rows,
        auto_grow=auto_grow,
        dense=dense,
        placeholder=placeholder if placeholder is not None else "",
        prefix=prefix if prefix is not None else "",
        suffix=suffix if suffix is not None else "",
    )
    use_change(text_area, set_value_cast, enabled=not continuous_update, update_events=update_events)
    return text_area



import dataclasses
import difflib
import importlib
import inspect
import pkgutil
import re
import linecache
import warnings
from pathlib import Path
from types import ModuleType
from typing import Any, Callable, Dict, List, Optional, cast

import ipywidgets
import reacton
import reacton.core

import solara
import solara.checks
from solara.alias import rv
from solara.util import cwd, nested_get

autoroute_level_context = solara.create_context(0)
DEBUG = False

solara_root = Path(solara.__file__).parent

DefaultLayout = solara.AppLayout
_redirects: Dict[str, str] = {}


def source_to_module(path: Path, initial_namespace={}) -> ModuleType:
    fullname = path.stem
    mod = ModuleType(fullname)
    mod.__file__ = str(path)
    mod.__dict__.update(initial_namespace)
    if path.suffix == ".py":
        with open(path, encoding="utf8") as f:
            ast = compile(f.read(), path, "exec")
            exec(ast, mod.__dict__)
    elif path.suffix == ".ipynb":
        # TODO: this is a dependency of solara-server, but this packages is in solara-ui
        import nbformat

        nb: nbformat.NotebookNode = nbformat.read(path, 4)
        with cwd(Path(path).parent):
            for cell_index, cell in enumerate(nb.cells):
                cell_index += 1  # used 1 based
                if cell.cell_type == "code":
                    source = cell.source
                    # similar to ipython/ipython/IPython/core/compilerop.py
                    # inspect.py:findsource accepts non-filenames with '<' and '>' around it
                    cell_path = f"<{str(path)} input cell {cell_index}>"
                    # put an entry in linecache so that we can use inspect.getsource
                    entry = (
                        len(source),
                        None,
                        [line + "\n" for line in source.splitlines()],
                        cell_path,
                    )
                    linecache.cache[cell_path] = entry
                    ast = compile(source, cell_path, "exec")
                    exec(ast, mod.__dict__)
    else:
        raise TypeError(f"Only .py and .ipynb supported (not {path})")
    return mod


def count_arguments(f: Callable):
    if isinstance(f, reacton.core.ComponentFunction):
        f = f.f
    sig = inspect.signature(f)
    return len([k for name, k in sig.parameters.items() if (k.default == k.empty) and (k.kind != k.VAR_POSITIONAL) and (k.kind != k.VAR_KEYWORD)])


def arg_cast(args: List[str], f: Callable):
    if isinstance(f, reacton.core.ComponentFunction):
        f = f.f
    sig = inspect.signature(f)
    results = []
    for i, (name, param) in enumerate(sig.parameters.items()):
        if i >= len(args):
            break
            # TODO: can we give a better error message?
            # raise TypeError(f"{f} expected positional argument {param}, but not enough argument supplied")
        annotation = param.annotation
        value = args[i]
        check_optional_types = [str, int, float]
        for check_type in check_optional_types:
            if annotation == Optional[check_type]:
                annotation = check_type
        if annotation == param.empty:
            results.append(value)
        else:
            try:
                results.append(annotation(value))
            except Exception as e:
                raise ValueError(f"Cannot cast argument {name} of {f} to type {annotation} with value {value}") from e
    return results


@solara.component
def RoutingProvider(children: List[reacton.core.Element] = [], routes: List[solara.Route] = [], pathname: str = ""):
    """Wraps the app, adds extra context, like navigation/routing."""
    path, set_path = solara.use_state(pathname, key="solara-context-path")

    def set_path_with_redirect(path):
        path = _redirects.get(path, path)
        set_path(path)

    # TODO: since we provide a cross filter context here, I don't think name `RoutingProvider` is a good name
    # we might want to change/refactor this.
    solara.provide_cross_filter()
    nav = solara.Navigator(location=path, on_location=set_path_with_redirect)
    solara.routing._location_context.provide(solara.routing._Location(path, set_path_with_redirect))
    solara.routing.router_context.provide(solara.routing.Router(path, routes=routes, set_path=set_path_with_redirect))

    def get_nav_widget():
        # not sure why get_widget(nav) does not work
        nav_widget.current = solara.get_widget(main).children[0]  # type: ignore

    import solara.widgets as w

    nav_widget = solara.use_ref(cast(Optional[w.Navigator], None))
    if nav_widget.current:
        nav_widget.current.location = path
    solara.use_effect(get_nav_widget)

    if solara.checks.should_perform_solara_check():
        children = [solara.checks.SolaraCheck(), *children]

    main = solara.VBox(
        children=[
            nav,
            *children,
        ]
    )

    return main


@solara.component
def RenderPage(main_name: str = "Page"):
    """Renders the page that matches the route."""
    router = solara.use_context(solara.routing.router_context)
    # we use these to cache script runs that use regular ipywidgets
    modules = cast(Dict[str, ModuleType], solara.use_memo(dict, dependencies=[]))
    modules_modified_times = cast(Dict[str, float], solara.use_memo(dict, dependencies=[]))

    if not router.path_routes:
        with solara.VBox() as main:
            solara.Error(f"Page not found: {router.path}")
            parent = "/" + "/".join(router.parts[:-1])
            with solara.Link(parent):
                solara.Button(f"Go to parent: {parent}", text=True)
            if DEBUG:
                from reacton.core import pp

                from solara.components.captureoutput import CaptureOutput

                with CaptureOutput():
                    pp(router.routes)
        return main

    layouts = []  # nested layouts
    level_max = len(router.path_routes) - 1
    for level in range(len(router.path_routes)):
        # we always level_start the 'package'/'root' layout
        roots = [k for k in router.path_routes_siblings[level] if k.path == "/"]
        if len(roots) == 1:
            root = roots[0]
            if root.layout:
                layouts.append(root.layout)
        route = router.path_routes[level]
        # and, if not the root layout, include the layout for the current route
        if route.path != "/" and route.layout:
            layouts.append(route.layout)
        # We found the leaf (or node) that defines a component or markdown file
        if (route.file and route.file.suffix == ".md") or route.component:
            level_max = level
            break

    route_current = router.path_routes[level_max]
    routes_siblings = router.path_routes_siblings[level_max]
    routes_siblings_index = routes_siblings.index(route_current)

    # if no layouts are found, we use the default layout
    if layouts == []:
        layouts = [DefaultLayout]

    if route_current.component is None and route_current.module is None and (route_current.file is None or route_current.file.suffix != ".md"):
        return solara.Error(f"Page not found: {router.path}, route does not link to a (markdown) path or module or component")

    def wrap_in_layouts(element: reacton.core.Element, layouts):
        for Layout in reversed(layouts):
            element = Layout(children=[element])
        return element

    def get_args(f):
        if len(router.path_routes) < len(router.parts):
            arg_strings = router.parts[len(router.path_routes) :]
            args = arg_cast(arg_strings, f)
            return args
        else:
            return []

    if str(route_current.file).endswith(".md") or isinstance(route_current.data, Path):
        path = cast(Path, route_current.data)
        if path.suffix == ".md":
            component = route_current.component or solara.Markdown
            with solara.HBox() as navigation:
                if routes_siblings_index > 0:
                    prev = routes_siblings[routes_siblings_index - 1]
                    with solara.Link(prev):
                        solara.Button(f"{prev.label}", text=True, icon_name="mdi-arrow-left")
                rv.Spacer()
                if routes_siblings_index < len(routes_siblings) - 1:
                    next = routes_siblings[routes_siblings_index + 1]
                    with solara.Link(next):
                        solara.Button(f"{next.label}", text=True, icon_name="mdi-arrow-right")
            is_relative = str(path.resolve()).startswith(str(solara_root.resolve()))
            if is_relative:
                # this is now hardcoded to solara only
                url = solara.util.github_edit_url(path)
                with solara.Div(style={"position": "relative"}) as content:
                    with solara.Tooltip("Edit on GitHub"):
                        solara.Button(
                            icon_name="mdi-pencil", icon=True, href=url, target="_blank", style={"position": "absolute", "top": "0px", "right": "0px"}
                        )
                    component(path.read_text(encoding="utf-8"), unsafe_solara_execute=True)
            else:
                content = component(path.read_text(encoding="utf-8"), unsafe_solara_execute=True)

            main = solara.Div(
                classes=["solara-autorouter-content"],
                children=[
                    solara.Title(route_current.label or "No title"),
                    content,
                    navigation,
                ],
            )
            main = wrap_in_layouts(main, layouts)
        else:
            main = solara.Error(f"Suffix {path.suffix} not supported")
    else:
        title = route_current.label or "No title"
        title_element = solara.Title(title)
        module = None
        Page = route_current.component
        if route_current.module is not None and (Page is None):
            # if not a custom component is given, we try to find a Page component
            # in the module
            assert route_current.module is not None
            module = route_current.module
            namespace = module.__dict__
            Page = nested_get(namespace, main_name, Page)
            if Page is None:
                # app is for backwards compatibility
                Page = namespace.get("page", namespace.get("app", Page))
                Page = nested_get(namespace, main_name, Page)

        if isinstance(Page, ipywidgets.Widget):
            # If we have a widget, we need to execute this again for each
            # connection, since we cannot share widgets between connections/users.
            # We also cannot tear them down, so we cache the widget based pages.
            # To support hot reload, we manually need to check the mtimes
            # because the reload support for modules in reloader.py only works
            # for modules.
            if route_current.file is None:
                page = solara.Error(f"{route_current.path} is not associated with a file")
            else:
                assert route_current.file is not None
                if route_current.path not in modules:
                    modules[route_current.path] = source_to_module(route_current.file)
                    modules_modified_times[route_current.path] = route_current.file.stat().st_mtime
                else:
                    if modules_modified_times[route_current.path] != route_current.file.stat().st_mtime:
                        # out of date, 'reload'
                        modules[route_current.path] = source_to_module(route_current.file)
                        modules_modified_times[route_current.path] = route_current.file.stat().st_mtime
                Page = nested_get(modules[route_current.path], main_name, None)
                if Page is None:
                    Page = getattr(modules[route_current.path], "app", None)
                    Page = getattr(modules[route_current.path], "page", Page)
            main = solara.Div(
                classes=["solara-autorouter-content"],
                children=[
                    title_element,
                    Page,
                ],
            )
            main = wrap_in_layouts(main, layouts)
        elif Page is not None:
            if isinstance(Page, reacton.core.ComponentFunction):
                args = get_args(Page)
                page = Page(*args)
            elif isinstance(Page, solara.Element):
                page = Page
            else:
                page = solara.Error(f"{Page} is not a component or element, but {type(Page)}")
            main = solara.Div(
                classes=["solara-autorouter-content"],
                children=[
                    title_element,
                    page,
                ],
            )
            main = wrap_in_layouts(main, layouts)
        else:
            if route_current.module:
                path_str = route_current.module.__file__
                local_scope = route_current.module.__dict__
                ignore = ["display"]
                options = [k for k in list(local_scope) if k not in ignore and not k.startswith("_")]
                matches = difflib.get_close_matches(main_name, options)
                msg = f"No object with name {main_name} found for {path_str}"
                if matches:
                    msg += " Did you mean: " + " or ".join(map(repr, matches))
                else:
                    msg += " We did find: " + " or ".join(map(repr, options))
            else:
                msg = f"{module} does not have a Page component or an app element"

            with DefaultLayout() as main:
                solara.Error(msg)
    return main


def get_page(module: ModuleType, required=True):
    page = getattr(module, "Page", None)
    if required and page is None:
        raise NameError(f"No component names 'Page' in module {module}")
    return page


def get_renderable(module: ModuleType, required=False):
    var_names = "app Page page".split()
    for var_name in var_names:
        if not hasattr(module, var_name):
            continue
        entry = getattr(module, var_name)
        return entry
    if required:
        raise NameError(f"No component, element or widget found in module {module} with names: {', '.join(var_names[:-1])} or {var_names[-1]}")


def get_title(module: ModuleType, required=True):
    assert module.__file__ is not None
    name = Path(module.__file__).stem
    if module.__file__.endswith("__init__.py"):
        name = Path(module.__file__).parent.stem
    # if title is a submodule it may shadown the variable in __init__
    # in this case, _title is an escape hatch
    if hasattr(module, "_title") and isinstance(module._title, str):
        title = module._title
    elif hasattr(module, "title") and isinstance(module.title, str):
        title = module.title
    else:
        match = re.match("([0-9\\-_ ]*)(.*)", name)
        if match:
            _prefix, name = match.groups()
        title_parts = re.split("[\\-_ ]+", name)
        title = " ".join(k.title() for k in title_parts)
    return title


def fix_route(route: solara.Route, new_file: Path, new_layout=None) -> solara.Route:
    file = route.file or new_file
    layout = route.layout or new_layout
    children = fix_routes(route.children, new_file) if route.children else []

    return dataclasses.replace(route, file=file, children=children, layout=layout)


def fix_routes(routes: List[solara.Route], new_file: Path, new_layout=None):
    new_routes = []
    for route in routes:
        if route.path == "/":
            route = fix_route(route, new_file, new_layout)
        else:
            route = fix_route(route, new_file)
        new_routes.append(route)
    return new_routes


def generate_routes(module: ModuleType) -> List[solara.Route]:
    """Generate routes from a module.

    This is a recursive function that will generate routes for all submodules.

    Note this only support .py files, since Markdown files and Jupyter notebook
    files are not Python modules.


    See [our multipage guide](/documentation/advanced/howto/multipage#as-a-package) for more details.


    """
    from .server import reload

    assert module.__file__ is not None
    routes = []
    children: List[solara.Route]
    file = Path(module.__file__)

    if module.__file__.endswith("__init__.py"):
        if hasattr(module, "routes"):
            # if routes if provided, use them instead of us generating them
            children = getattr(module, "routes")
            children = fix_routes(children, file)
            return children
        route_order = getattr(module, "route_order", None)
        layout = getattr(module, "Layout", None)
        title = get_title(module)
        children = getattr(module, "routes", [])
        if hasattr(module, "Page"):
            routes.append(
                solara.Route(path="/", component=get_page(module), data=module, module=module, layout=layout, children=children, label=title, file=file)
            )

        assert module.__file__ is not None
        reload.reloader.watcher.add_file(module.__file__)
        for info in pkgutil.iter_modules([str(Path(module.__file__).parent)]):
            submod = importlib.import_module(module.__name__ + f".{info.name}")
            title = get_title(submod)

            name = info.name
            # ideally, we do this similar to generate_routes_directory
            # however, this may break things.
            # name = name.replace("_", "-")
            if info.ispkg:
                # we are in a package, like 'portal/solara_portal/pages', not the module itself
                # (e.g. portal/solara_portal/pages/__init__.py)
                # so here name='pages', and the children will come from the submodules
                children = generate_routes(submod)
                route = solara.Route(name, component=None, children=children, module=submod, layout=None, label=title)
                # skip empty subpackages
                if len(route.children) == 0:
                    continue
            else:
                # skip empty modules
                if get_renderable(submod) is None and not hasattr(submod, "routes"):
                    continue
                route = _generate_route(name, submod)
            routes.append(route)
        if route_order:
            lookup = {k.path: k for k in routes}
            for k in route_order:
                if k not in lookup:
                    raise KeyError(f"Route {k!r} listen in route_order not found in {module}")
            routes = [lookup[k] for k in route_order]
            if set(lookup) - set(route_order):
                warnings.warn(f"Some routes are not in route_order: {set(lookup) - set(route_order)}")

    else:
        return [_generate_route("/", module)]

    return routes


def generate_routes_directory(path: Path, markdown_component=None) -> List[solara.Route]:
    """Generate routes for a directory.

    This is a recursive function that will generate routes for all
    subdirectories and files in the directory. It will skip any
    files or directories that start with an underscore or a dot.

    Markdown files ending in .md will be rendered as markdown.

    Python files ending in .py, or Jupyter notebooks ending in .ipynb
    will be executed and its `Page` component will be rendered.

    Automatic titles will be [generated as explained in the multipage guide](/documentation/advanced/howto/multipage).

    """

    subpaths = list(sorted(path.iterdir()))
    routes = []
    first = True
    has_index = len([k for k in subpaths if k.name == "__init__"]) > 0
    init = path / "__init__.py"
    layout = None
    if init.exists():
        init_module = source_to_module(init)
        layout = getattr(init_module, "Layout", None)

    suffixes = [".py", ".ipynb", ".md"]
    for subpath in subpaths:
        # only handle directories and recognized file types
        if not (subpath.is_dir() or subpath.suffix in suffixes):
            continue
        if subpath.stem.startswith("_") or subpath.stem.startswith("."):
            continue
        route = _generate_route_path(subpath, layout=layout, first=first, has_index=has_index, markdown_component=markdown_component)
        first = False
        routes.append(route)
    return routes


def _generate_route_path(subpath: Path, layout=None, first=False, has_index=False, initial_namespace={}, markdown_component=None) -> solara.Route:
    from .server import reload

    name = subpath.stem
    match = re.match("([0-9\\-_ ]*)(.*)", name)
    if match:
        _prefix, name = match.groups()
    title_parts = re.split("[\\-_ ]+", name)
    title = " ".join(k.title() for k in title_parts)
    if not has_index and first:
        route_path = "/"
    else:
        route_path = "-".join([k.lower() for k in title_parts])
    component = None
    children: List[solara.Route] = []
    module: Optional[ModuleType] = None
    data: Any = None
    module_layout = layout if first else None
    if subpath.suffix == ".md":
        data = subpath
        component = markdown_component
        reload.reloader.watcher.add_file(subpath)
    elif subpath.is_dir():
        children = generate_routes_directory(subpath, markdown_component=markdown_component)
    else:
        reload.reloader.watcher.add_file(subpath)
        module = source_to_module(subpath, initial_namespace=initial_namespace)
        return _generate_route(route_path, module, default_layout=module_layout)
    route = solara.Route(route_path, component=component, module=module, label=title, children=children, data=data, layout=layout, file=subpath)
    return route


def _generate_route(route_path: str, module: ModuleType, default_layout=None, data=None) -> solara.Route:
    path = Path(module.__file__) if module.__file__ is not None else None
    title = get_title(module)
    layout = getattr(module, "Layout", default_layout)
    if inspect.isclass(layout) and issubclass(layout, ipywidgets.Layout):
        layout = None
    component = None
    children = getattr(module, "routes", [])
    root_route = get_root(children)
    # if we have no children or the children have no explicit root component (at '/')
    if not children or (root_route and root_route.component is None):
        component = get_page(module, required=False)
    if root_route and component and root_route.component and component is not root_route.component:
        warnings.warn(
            f"In {path}, you defined a Page component, but also a component on the root route (with path='/') "
            "which is not equal to the Page component at the module level. This is not recommended."
        )
    return solara.Route(route_path, component=component, module=module, label=title, children=children, data=data, layout=layout, file=path)


def get_root(routes: List[solara.Route]) -> Optional[solara.Route]:
    for route in routes:
        if route.path == "/":
            return route
    return None



import json
import logging
import pdb
import queue
import re
import struct
import warnings
from binascii import b2a_base64
from datetime import datetime
from typing import Set, Union

import ipykernel
import ipykernel.kernelbase
import jupyter_client.session as session
from dateutil.tz import tzlocal  # type: ignore
from ipykernel.comm import CommManager
from zmq.eventloop.zmqstream import ZMQStream

import solara
from solara.server.shell import SolaraInteractiveShell

from . import settings, websocket

logger = logging.getLogger("solara.server.kernel")
ipykernel_major = int(ipykernel.__version__.split(".")[0])

jsonmodule = json


# from jupyter_client/jsonutil.py
def _ensure_tzinfo(dt: datetime) -> datetime:
    """Ensure a datetime object has tzinfo
    If no tzinfo is present, add tzlocal
    """
    if not dt.tzinfo:
        # No more naïve datetime objects!
        warnings.warn(
            "Interpreting naive datetime as local %s. Please add timezone info to timestamps." % dt,
            DeprecationWarning,
            stacklevel=4,
        )
        dt = dt.replace(tzinfo=tzlocal())
    return dt


def json_default(obj):
    """default function for packing objects in JSON."""
    if isinstance(obj, datetime):
        obj = _ensure_tzinfo(obj)
        return obj.isoformat().replace("+00:00", "Z")
    elif isinstance(obj, bytes):
        return b2a_base64(obj).decode("ascii")
    if type(obj).__module__ == "numpy":
        import numpy as np

        if isinstance(obj, np.number):
            return obj.item()
        else:
            raise TypeError("%r is not JSON serializable" % obj)
    else:
        raise TypeError("%r is not JSON serializable" % obj)


def json_dumps(data):
    try:
        return jsonmodule.dumps(data)
    except TypeError:
        logger.warning("Invalid JSON, will try a with a more forgiving json encoder")
    return jsonmodule.dumps(
        data,
        default=json_default,
        ensure_ascii=False,
        allow_nan=False,
    )


ipykernel_version = tuple(map(int, re.split(r"\D+", ipykernel.__version__)[:3]))
if ipykernel_version >= (6, 18, 0):
    import comm.base_comm

    class Comm(comm.base_comm.BaseComm):
        kernel: Union[ipykernel.kernelbase.Kernel, None]

        def __init__(self, **kwargs) -> None:
            if ipykernel.kernelbase.Kernel.initialized():
                self.kernel = ipykernel.kernelbase.Kernel.instance()
            else:
                self.kernel = None
            super().__init__(**kwargs)

        def publish_msg(self, msg_type, data=None, metadata=None, buffers=None, **keys):
            if self.kernel is None or self.kernel.session is None:
                return
            data = {} if data is None else data
            metadata = {} if metadata is None else metadata
            content = dict(data=data, comm_id=self.comm_id, **keys)
            self.kernel.session.send(
                self.kernel.iopub_socket,
                msg_type,
                content,
                metadata=metadata,
                parent=self.kernel.get_parent("shell"),
                ident=self.topic,
                buffers=buffers,
            )

    comm.create_comm = Comm

    def get_comm_manager():
        from .kernel_context import get_current_context, has_current_context

        if has_current_context():
            return get_current_context().kernel.comm_manager
        else:
            return global_comm_manager

    global_comm_manager = comm.get_comm_manager()
    comm.get_comm_manager = get_comm_manager

# from notebook.base.zmqhandlers import serialize_binary_message
# this saves us a dependency on notebook/jupyter_server when e.g.
# running on pyodide


def _fix_msg(msg):
    # makes sure the msg can be json serializable
    # instead of using a callable like in jupyter_client (i.e. json_default)
    # we replace the keys we know are problematic
    # this allows us to use a faster json serializer in the future
    if "header" in msg and "date" in msg["header"]:
        # this is what jupyter_client.jsonutil.json_default does
        msg["header"]["date"] = msg["header"]["date"].isoformat().replace("+00:00", "Z")
    if "parent_header" in msg and "date" in msg["parent_header"]:
        # date is already a string if it's copied from the header that is not turned into a datetime
        # maybe we should do that in server.py
        date = msg["parent_header"]["date"]
        if isinstance(date, datetime):
            msg["parent_header"]["date"] = date.isoformat().replace("+00:00", "Z")


def serialize_binary_message(msg):
    """serialize a message as a binary blob

    Header:

    4 bytes: number of msg parts (nbufs) as 32b int
    4 * nbufs bytes: offset for each buffer as integer as 32b int

    Offsets are from the start of the buffer, including the header.

    Returns
    -------

    The message serialized to bytes.

    """
    # don't modify msg or buffer list in-place
    msg = msg.copy()
    buffers = list(msg.pop("buffers"))
    bmsg = json_dumps(msg).encode("utf8")
    buffers.insert(0, bmsg)
    nbufs = len(buffers)
    offsets = [4 * (nbufs + 1)]
    for buf in buffers[:-1]:
        offsets.append(offsets[-1] + len(buf))
    offsets_buf = struct.pack("!" + "I" * (nbufs + 1), nbufs, *offsets)
    buffers.insert(0, offsets_buf)
    return b"".join(buffers)


def deserialize_binary_message(bmsg):
    """deserialize a message from a binary blog

    Header:

    4 bytes: number of msg parts (nbufs) as 32b int
    4 * nbufs bytes: offset for each buffer as integer as 32b int

    Offsets are from the start of the buffer, including the header.

    Returns
    -------
    message dictionary
    """
    nbufs = struct.unpack("!i", bmsg[:4])[0]
    offsets = list(struct.unpack("!" + "I" * nbufs, bmsg[4 : 4 * (nbufs + 1)]))
    offsets.append(None)
    bufs = []
    for start, stop in zip(offsets[:-1], offsets[1:]):
        bufs.append(bmsg[start:stop])
    msg = json.loads(bufs[0].decode("utf8"))
    msg["buffers"] = bufs[1:]
    return msg


SESSION_KEY = b"solara"


class WebsocketStream:
    def __init__(self, session, channel: str):
        self.session = session
        self.channel = channel


class WebsocketStreamWrapper(ZMQStream):
    def __init__(self, websocket, channel):
        self.websocket = websocket
        self.channel = channel

    def flush(self, *ignore):
        pass


def send_websockets(websockets: Set[websocket.WebsocketWrapper], binary_msg):
    for ws in list(websockets):
        try:
            ws.send(binary_msg)
        except websocket.WebSocketDisconnect:
            # ignore the exception, we tried to send while websocket closed
            # just remove it from the websocket set
            try:
                # websocket can be modified by another thread
                websockets.remove(ws)
            except KeyError:
                pass  # already removed
        except Exception as e:  # noqa
            logger.exception("Error sending message: %s, closing websocket", e)
            try:
                ws.close()
            except websocket.WebSocketDisconnect:
                pass  # was already closed, which triggered the original exception
            except Exception as e2:  # noqa
                logger.exception("Error closing websocket: %s %s", e, e2)
            try:
                # websocket can be modified by another thread
                websockets.remove(ws)
            except KeyError:
                pass  # already removed


class SessionWebsocket(session.Session):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.websockets: Set[websocket.WebsocketWrapper] = set()  # map from .. msg id to websocket?

    def close(self):
        for ws in list(self.websockets):
            try:
                ws.close()
            except:  # noqa
                pass

    def send(
        self,
        stream,
        msg_or_type,
        content=None,
        parent=None,
        ident=None,
        buffers=None,
        track=False,
        header=None,
        metadata=None,
    ):
        if stream is None:
            return  # can happen when the kernel is closed but someone was still trying to send a message
        try:
            if isinstance(msg_or_type, dict):
                msg = msg_or_type
            else:
                msg = self.msg(
                    msg_or_type,
                    content=content,
                    parent=parent,
                    header=header,
                    metadata=metadata,
                )
            _fix_msg(msg)
            msg["channel"] = stream.channel
            # not using pdb guard for performance reasons
            try:
                if buffers:
                    msg["buffers"] = [memoryview(k).cast("b") for k in buffers]
                    wire_message = serialize_binary_message(msg)
                else:
                    wire_message = json_dumps(msg)
            except Exception:
                logger.exception("Could not serialize message: %r", msg)
                if settings.main.use_pdb:
                    pdb.post_mortem()
                raise
            send_websockets(self.websockets, wire_message)
        except Exception as e:
            logger.exception("Error sending message: %s", e)


class Kernel(ipykernel.kernelbase.Kernel):
    session: SessionWebsocket  # type: ignore
    # Ideally we have `session = Instance(Session, allow_none=True)`, but MyPy does not like it

    implementation = "solara"
    implementation_version = solara.__version__
    banner = "solara"

    def __init__(self):
        super().__init__()
        self.session = SessionWebsocket(parent=self, key=SESSION_KEY)
        self.msg_queue = queue.Queue()  # type: ignore
        self.stream = self.iopub_socket = WebsocketStream(self.session, "iopub")
        # on github action the next line gives a mypy error:
        # solara/server/kernel.py:111: error: "SessionWebsocket" has no attribute "stream"
        # not sure why we cannot reproduce that locally
        self.session.stream = self.iopub_socket  # type: ignore
        if ipykernel_version >= (6, 18, 0):
            # from this version on, ipykernel uses the comm package https://github.com/ipython/ipykernel/pull/973
            self.comm_manager = CommManager(parent=self, kernel=self)
            import ipywidgets.widgets.widget

            if hasattr(ipywidgets.widgets.widget, "Comm"):
                ipywidgets.widgets.widget.Comm = Comm
                ipywidgets.widgets.widget.Widget.comm.klass = Comm
        else:
            self.comm_manager = CommManager(parent=self, kernel=self)
        self.log = logging.getLogger("fake")

        comm_msg_types = ["comm_open", "comm_msg", "comm_close"]
        for msg_type in comm_msg_types:
            self.shell_handlers[msg_type] = getattr(self.comm_manager, msg_type)
        self.shell = SolaraInteractiveShell()
        self.shell.display_pub.session = self.session
        self.shell.display_pub.pub_socket = self.iopub_socket

    def close(self):
        if self.comm_manager is None:
            raise RuntimeError("Kernel already closed")
        self.session.close()
        self._cleanup_references()

    def _cleanup_references(self):
        try:
            # all of these reduce the circular references
            # making it easier for the garbage collector to clean up
            self.shell_handlers.clear()
            self.control_handlers.clear()
            for comm_object in list(self.comm_manager.comms.values()):  # type: ignore
                comm_object.close()
            self.comm_manager.targets.clear()  # type: ignore
            # self.comm_manager.kernel points to us, but we cannot set it to None
            # so we remove the circular reference by setting the comm_manager to None
            self.comm_manager = None  # type: ignore
            self.session.parent = None  # type: ignore

            self.shell.display_pub.session = None  # type: ignore
            self.shell.display_pub.pub_socket = None  # type: ignore
            del self.shell.__dict__
            self.shell = None  # type: ignore
            self.session.websockets.clear()
            self.session.stream = None  # type: ignore
            self.session = None  # type: ignore
            self.stream.session = None  # type: ignore
            self.stream = None  # type: ignore
            self.iopub_socket = None  # type: ignore
        except Exception:
            logger.exception("Error cleaning up references from kernel, not fatal")

    async def _flush_control_queue(self):
        pass

    # these don't work from non-main thread, and we do not care about them I think
    # TODO: it seems that if post_handler_hook is not override, the flask reload tests fails
    # for unknown reason
    def pre_handler_hook(self, *args):
        pass

    def post_handler_hook(self, *args):
        pass

    def set_parent(self, ident, parent, channel="shell"):
        """Overridden from parent to tell the display hook and output streams
        about the parent message.
        """
        if ipykernel_major < 6:
            # the channel argument was added in 6.0
            super().set_parent(ident, parent)
        else:
            super().set_parent(ident, parent, channel)
        if channel == "shell":
            self.shell.set_parent(parent)
